// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund.Components.Climate;

interface

uses Esmf;

type
  /// <summary>Simple climate component</summary>
  ClimateComponent = public class
  private
    /// <summary>interaction radiative forcing methane and nitrous oxide (in ppbv)</summary>
    method Interact(M,N: double): double;
  public
    /// <summary>Calculates all climate variables</summary>
    method Run(clock: Clock; state: IClimateState);
  end;
  
implementation

method ClimateComponent.Interact(M,N: double): double;
var
  hulp: double;
begin
  hulp:= 1 + Math.Pow(M*N,0.75)*2.01E-5 + Math.Pow(M*N,1.52)*M*5.31E-15;
  Result:= 0.47*Math.Log(hulp)
end;

method ClimateComponent.Run(clock: Clock; state: IClimateState);
begin
  // create shortcuts for commonly accessed data
  var s:=state;
  var t:=clock.Current;
  
  // 
  var CH4N2O: double :=Interact(s.ch4pre,s.n2opre);
              
  // Calculate CO2 concentrations
  s.cbox1[t]:=s.cbox1[t-1]*s.co2decay1 + 0.000471*s.co2frac1*(s.mco2[t] + iif(s.addmp,s.addco2[t],0));
  s.cbox2[t]:=s.cbox2[t-1]*s.co2decay2 + 0.000471*s.co2frac2*(s.mco2[t] + iif(s.addmp,s.addco2[t],0));
  s.cbox3[t]:=s.cbox3[t-1]*s.co2decay3 + 0.000471*s.co2frac3*(s.mco2[t] + iif(s.addmp,s.addco2[t],0));
  s.cbox4[t]:=s.cbox4[t-1]*s.co2decay4 + 0.000471*s.co2frac4*(s.mco2[t] + iif(s.addmp,s.addco2[t],0));
  s.cbox5[t]:=s.cbox5[t-1]*s.co2decay5 + 0.000471*s.co2frac5*(s.mco2[t] + iif(s.addmp,s.addco2[t],0));

  s.acco2[t]:=s.cbox1[t]+s.cbox2[t]+s.cbox3[t]+s.cbox4[t]+s.cbox5[t];
    
  // Calculate CH4 concentrations
  var globch4:=s.globch4[t] + iif(s.addmp,s.addch4[t],0) + s.RunAwayCH4[t] + iif(t.Value>50,s.ch4add*(t.Value-50),0);
  if globch4<0 then
    globch4:=0;
  s.acch4[t]:=s.acch4[t-1] + 0.3597*globch4 - s.ch4decay*(s.acch4[t-1]-s.ch4pre);
  
  if s.acch4[t]<0 then
    raise new ApplicationException('ch4 atmospheric concentration out of range');
    
  // Calculate N2O concentrations
  var globn2o:=s.globn2o[t] + iif(s.addmp,s.addn2o[t],0) + iif(t.Value>50, s.n2oadd*(t.Value-50),0);
  s.acn2o[t]:=s.acn2o[t-1] + 0.2079*globn2o - s.n2odecay*(s.acn2o[t-1]-s.n2opre);
  
  if s.acn2o[t]<0 then
    raise new ApplicationException('n2o atmospheric concentration out of range');
           
  // Calculate SF6 concentrations   
  var globsf6:=s.globsf6[t] + iif(s.addmp,s.addsf6[t],0) + iif(t.Value>50, s.sf6add*(t.Value-50),0);
  if globsf6<0 then
    globsf6:=0;  
  s.acsf6[t]:=s.sf6pre + (s.acsf6[t-1]-s.sf6pre)*(1-s.sf6decay) + globsf6/25.1;
  
  if s.acsf6[t]<0 then
    raise new ApplicationException('sf6 atmospheric concentration out of range');
              
  // Calculate SO2 concentrations
  s.acso2[t]:=s.globso2[t];  

  // Calculate radiative forcing        
                         
  s.rfCO2[t]:=5.35*Math.Log(s.acco2[t]/s.co2pre);
  
  s.rfCH4[t]:=0.036*(Math.Sqrt(s.acCH4[t])-Math.Sqrt(s.ch4pre))
    - interact(s.acch4[t],s.n2opre) + CH4N2O;
                     
  s.rfN2O[t]:=0.14*(Math.Sqrt(s.acN2O[t])-Math.Sqrt(s.n2opre))
    - interact(s.ch4pre,s.acn2o[t]) + CH4N2O;
    
  s.rfSF6[t]:=0.00052*(s.acsf6[t]-s.sf6pre);
  
  s.rfSO2[t]:=s.so2dir*s.acso2[t]/14.6 + s.so2ind*Math.Log(1+s.acso2[t]/34.4)/Math.Log(1+14.6/34.4) - 0.9;
  // s.rfSO2[t]:=s.so2dir*s.globso2[t]/14.6 + s.so2ind*ln(1+s.globso2[t]/34.4)/ln(1+14.6/34.4) - 0.9;
  
  s.radforc[t]:=s.rfCO2[t] + s.rfCH4[t] + s.rfN2O[t] + s.rfsf6[t] - s.rfSO2[t];
  
  if s.radforc[t] > s.maxrf THEN
     s.maxrf:=s.radforc[t];

     
  // Calculate temperature          
  var dtemp:=s.delaytemp*s.temps*s.radforc[t] - s.delaytemp*s.temp[t-1];
          
  s.temp[t]:=s.temp[t-1] + dtemp;
          
end;

end.