// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund.Components.HotellingTaxation;

interface

uses Esmf,
     Fund.CommonDimensions;

type
  
  IGlobalHotellingTaxationState = public interface
    property basetax: double read;
    property baseyear: Timestep read;
    
    property prtp: Double read;
    property elasticityofmarginalutility: Double read;    
    
    property population: IParameter1Dimensional<Timestep, double> read;
    property consumption: IParameter1Dimensional<Timestep, double> read;
    
    property currtax: IVariable2Dimensional<Timestep, Region, double> read;        
  end;
  
  IRegionalHotellingTaxationState = public interface
    property basetax: IParameter1Dimensional<Region, double> read;
    property baseyear: Timestep read;
    
    property prtp: Double read;
    property elasticityofmarginalutility: Double read;    
    
    property population: IParameter2Dimensional<Timestep, Region, double> read;
    property consumption: IParameter2Dimensional<Timestep, Region, double> read;
    
    property currtax: IVariable2Dimensional<Timestep, Region, double> read;        
  end;
  
  
  GlobalHotellingTaxationComponent = public class
  private
  protected
  public
    method Run(clock: Clock; state: IGlobalHotellingTaxationState);    
  end;
  
  RegionalHotellingTaxationComponent = public class
  private
  protected
  public
    method Run(clock: Clock; state: IRegionalHotellingTaxationState);    
  end;  
  
implementation

method GlobalHotellingTaxationComponent.Run(clock: Clock; state: IGlobalHotellingTaxationState); 
begin
  var t:=clock.Current;
  var s:=state;
  
  var perCapitaConsumptionNow:=s.consumption[t]/s.population[t];
  var perCapitaConsumptionPrevious:=s.consumption[t-1]/s.population[t-1];
  
  var perCapitaGrowthRate:=perCapitaConsumptionNow/perCapitaConsumptionPrevious - 1.0;
  
  var discountrate:=perCapitaGrowthRate * s.elasticityofmarginalutility + s.prtp;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    if t<s.baseyear then
      s.currtax[t,r]:=0.0
    else if t=s.baseyear then
      s.currtax[t,r]:=s.basetax
    else
    begin
      s.currtax[t,r]:= s.currtax[t-1,r] * (1.0 + discountrate);
    end;      
  end;
  
end;

method RegionalHotellingTaxationComponent.Run(clock: Clock; state: IRegionalHotellingTaxationState); 
begin
  var t:=clock.Current;
  var s:=state;  
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    var perCapitaConsumptionNow:=s.consumption[t,r]/s.population[t,r];
    var perCapitaConsumptionPrevious:=s.consumption[t-1,r]/s.population[t-1,r];
    
    var perCapitaGrowthRate:=perCapitaConsumptionNow/perCapitaConsumptionPrevious - 1.0;
    
    var discountrate:=perCapitaGrowthRate * s.elasticityofmarginalutility + s.prtp;
    
    if t<s.baseyear then
      s.currtax[t,r]:=0.0
    else if t=s.baseyear then
      s.currtax[t,r]:=s.basetax[r]
    else
    begin
      s.currtax[t,r]:= s.currtax[t-1,r] * (1.0 + discountrate);
    end;      
  end;
  
end;

end.