// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund.Components.ImpactAgriculture;

interface

uses Esmf,
     Fund.CommonDimensions;

type
  
  IImpactAgricultureState = public interface
    property gdp90: IParameter1Dimensional<Region, double> read;
    property income: IParameter2Dimensional<Timestep, Region, double> read;
    property pop90: IParameter1Dimensional<Region, double> read;    
    property population: IParameter2Dimensional<Timestep, Region, double> read;
       
    property agrish: IVariable2Dimensional<Timestep, Region, double> read;
    property agrish0: IParameter1Dimensional<Region, double> read;
    property agel: Double read;    
    
    property agrate: IVariable2Dimensional<Timestep, Region, double> read;
    property aglevel: IVariable2Dimensional<Timestep, Region, double> read;
    property agco2: IVariable2Dimensional<Timestep, Region, double> read;
    property agcost: IVariable2Dimensional<Timestep, Region, double> read;
    
    property agrbm: IParameter1Dimensional<Region, double> read;
    property agtime: IParameter1Dimensional<Region, double> read;
    property agnl: Double read;    
       
    property aglbm: IParameter1Dimensional<Region, double> read;
    property agopt: IParameter1Dimensional<Region, double> read;
    
    property agcbm: IParameter1Dimensional<Region, double> read;
    property co2pre: Double read;        
    
    property temp: IParameter1Dimensional<Timestep, double> read;
    property acco2: IParameter1Dimensional<Timestep, double> read;        
  end;
  
  ImpactAgricultureComponent = public class
  private
  protected
    method Getaglparl(Region: Fund.CommonDimensions.Region; state: IImpactAgricultureState): double;    
    method Getaglparq(Region: Fund.CommonDimensions.Region; state: IImpactAgricultureState): double;    
  public
    method Run(clock: Clock; state: IImpactAgricultureState);            
  end;
  
implementation

method ImpactAgricultureComponent.Getaglparl(Region: Fund.CommonDimensions.Region; state: IImpactAgricultureState): double;
var
  i_aglparq: double;
begin
  var s:=state;
  
  i_aglparq:=s.aglbm[Region]/(10.24-6.4*s.agopt[Region]);
  if i_aglparq > 0 then
    Result:=s.aglbm[Region]/3.2
  else
    Result:=-2.0*s.agopt[Region]*s.aglbm[Region]/(10.24-6.4*s.agopt[Region]);
end;

method ImpactAgricultureComponent.Getaglparq(Region: Fund.CommonDimensions.Region; state: IImpactAgricultureState): double;
var
  i_aglparq: double;
begin
  var s:=state;
  
  i_aglparq:=s.aglbm[Region]/(10.24-6.4*s.agopt[Region]);
  if i_aglparq > 0 then
    Result:=0.0
  else
    Result:=i_aglparq;
end;

method ImpactAgricultureComponent.Run(clock: Clock; state: IImpactAgricultureState); 
begin
  var s:=state;
  var t:=clock.Current;
  
  var dtemp:=s.temp[t]-s.temp[t-1];

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin    
    var ypc: Double:=s.income[t,r]/s.population[t,r]*1000.0;
    var ypc90: Double:=s.gdp90[r]/s.pop90[r]*1000.0;
    
    s.agrish[t,r]:=s.agrish0[r]*Math.Pow(ypc/ypc90,-s.agel);
  end;

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    if double.IsNaN(Math.Pow(dtemp/0.04,s.agnl)) then
      s.agrate[t,r]:=0.0
    else
      s.agrate[t,r]:=s.agrbm[r]*Math.Pow(dtemp/0.04,s.agnl) + (1.0-1.0/s.agtime[r])*s.agrate[t-1,r];        
  end;  

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin    
    s.aglevel[t,r]:=Getaglparl(r,s)*s.temp[t] + Getaglparq(r,s)*Math.Pow(s.temp[t],2.0);
  end;              
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin    
    s.agco2[t,r]:=s.agcbm[r]/Math.Log(2.0)*Math.Log(s.acco2[t-1]/s.co2pre);
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    s.agcost[t,r]:=Math.Min(1.0,s.agrate[t,r]+s.aglevel[t,r]+s.agco2[t,r]) * s.agrish[t,r] * s.income[t,r];
  end;    
end;

end.