// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund.Components.ImpactBioDiversity;

interface

uses Esmf,
  Fund.CommonDimensions;

type
  
  IImpactBioDiversityState = public interface
    /// <summary> Change in number of species in relation to the year 2000 </summary>
    property biodiv: IVariable1Dimensional<Timestep, double> read;
       
    property species: IVariable2Dimensional<Timestep, Region, double> read;
    
    /// <summary> Number of species in the year 2000 </summary>
    property nospecbase: double read;
    
    property nospecies: IParameter1Dimensional<Timestep, double> read;
    property temp: IParameter1Dimensional<Timestep, double> read;
    property income: IParameter2Dimensional<Timestep, Region, double> read;
    property population: IParameter2Dimensional<Timestep, Region, double> read;
    property valinc: IParameter1Dimensional<Region, double> read;
    property bioshare: double read;
    property spbm: double read;
    property valbase: double read;
    property dbsta: double read;
  end;
  
  ImpactBioDiversityComponent = public class
  private
  protected
  public
    /// <summary>Calculates all biodiversity impacts</summary>
    method Run(clock: Clock; state: IImpactBioDiversityState);    
  end;
  
implementation

method ImpactBioDiversityComponent.Run(clock: Clock; state: IImpactBioDiversityState); 
begin
  var s:=state;
  var t:=clock.Current;

  s.biodiv[t]:=s.nospecbase/s.nospecies[t];

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin  
    var ypc:double:=1000.0*s.income[t,r]/s.population[t,r];
        
    var dt:=s.temp[t]-s.temp[t-1];

    var valadj:=s.valbase/s.valinc[r]/(1+s.valbase/s.valinc[r]);
        
    s.species[t,r]:=s.spbm/
                    s.valbase*ypc/s.valinc[r]/(1.0+ypc/s.valinc[r])/valadj*ypc*
                    s.population[t,r]/1000.0*
                    dt/s.dbsta/(1.0+dt/s.dbsta)*
                    (1.0-s.bioshare+s.bioshare*s.biodiv[t]);                    
  end;

end;

end.