// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund.Components.SocioEconomic;

interface

uses System.Collections.Generic,
  Esmf,
  Fund.CommonDimensions;

type
    
  SocioEconomicComponent = public class
  private
  protected
  public
    method Run(clock: Clock; state: ISocioEconomicState);    
  end;
  
implementation

method SocioEconomicComponent.Run(clock: Clock; state: ISocioEconomicState); 
begin
  var s:=state;
  var t:=clock.Current;
  
  var totalPopulation:=0.0;
  // Calculate population
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.population[t,r]:=(1 + 0.01*s.pgrowth[t-1, r])*(s.population[t-1,r] + iif( (t>=Timestep.FromSimulationYear(40)) and not s.runwithoutpopulationperturbation, s.enter[t-1,r] - s.leave[t-1,r] - iif(s.dead[t-1,r]>=0, s.dead[t-1,r],0), 0));
        
    if s.population[t,r]<0 then
      s.population[t,r]:=0.000001;
      //raise new Exception;
      
   s.populationin1[t,r]:=s.population[t,r]*1000000.0;
   totalPopulation:=totalPopulation+s.populationin1[t,r];
  end;

  s.globalpopulation[t]:=totalPopulation;  
  
  // Initialize data structure
  var ygrowth:=new Dictionary<Region,double>;  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    ygrowth.Add(r,0.0);
  end;

  // Calculate income growth rate  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    ygrowth[r]:=(1+0.01*s.pgrowth[t-1,r]) * (1+0.01*s.ypcgrowth[t-1,r]) - 1;
  end; 

  // Calculate income  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    var oldincome:double:=s.income[t-1,r] - iif( (t>=Timestep.FromSimulationYear(40)) and not s.runwithoutdamage,s.consleak*s.eloss[t-1,r]/10,0);
          
    s.income[t,r]:= (1 + ygrowth[r] - s.taxmp[r]*s.ryg[t-1,r] - s.ch4cost[t-1,r] - s.n2ocost[t-1,r])*oldincome - s.formp[r]*s.forcost[t-1,r]*oldincome/s.gdp0[r];
  end;
  
  // Check for unrealistic values
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    if s.income[t,r] < 0.1*s.population[t,r] then
    begin
      s.income[t,r]:=0.1*s.population[t,r];
    end;
  
  end;
  
  var totalConsumption:=0.0;
  var savingsrate: Double := 0.2;
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.consumption[t,r]:=Math.Max(
      s.income[t,r]*1000000000.0*(1.0-savingsrate)-iif(s.runwithoutdamage, 0.0, (s.eloss[t-1,r] + s.sloss[t-1,r])*1000000000.0),
      0.0);
      
    totalConsumption:=totalConsumption+s.consumption[t,r];
  end;
  s.globalconsumption[t]:=totalConsumption;
  
  // Calculate emission and carbon intensity
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.energint[t,r]:=(1.0 - 0.01*s.aeei[t,r] - s.reei[t-1,r])*s.energint[t-1,r];
    s.emissint[t,r]:=(1.0 - 0.01*s.acei[t,r] - s.rcei[t-1,r])*s.emissint[t-1,r];
  end;   
  
  // Calculate so2 emissions
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.so2[t,r]:= s.so2[t-1,r] * 
                Math.Pow(1+0.01*s.pgrowth[t-1,r],s.so2pop) *
                Math.Pow(1+0.01*s.ypcgrowth[t-1,r],s.so2inc) *
                Math.Pow(1 - 0.01*s.aeei[t,r] - s.reei[t-1,r] - 0.01*s.acei[t,r] - s.rcei[t-1,r], s.so2carb);
    
    s.so2WithSeeiAndScei[t,r]:=s.so2[t,r] * (1-s.seei[t-1,r]-s.scei[t-1,r]);
  end;
  
  // Calculate sf6 emissions
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin  
    s.sf6[t,r]:=(s.sf60[r] + s.sf6gdp*(s.income[t,r]-s.GDP90[r]) + s.sf6ypc*(s.income[t-1,r]/s.population[t-1,r]-s.GDP90[r]/s.pop90[r]))*iif(t<=Timestep.FromSimulationYear(60),1+(t.Value-40.0)/40.0,1.0+(60.0-40.0)/40.0)*iif(t>Timestep.FromSimulationYear(60), Math.Pow(0.99,t.Value-60.0), 1.0);
  end;
  
  // Check for unrealistic values
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    if s.sf6[t,r] < 0.0 then
    begin
      s.sf6[t,r]:=0;
    end;
  
  end;
  
  // Calucate energy use
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.energuse[t,r]:=(1-s.seei[t-1,r])*s.energint[t,r]*s.income[t,r];
  end;
  
  // Calculate co2 emissions
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.emission[t,r]:=(1-s.scei[t-1,r])*s.emissint[t,r]*s.energuse[t,r];
    s.emissionwithforestry[t,r]:=s.emission[t,r] + s.forestemm[t,r] - s.forestry[t-1,r];
  end;

  // Calculate ch4 emissions
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.ch4[t,r]:=s.ch4em[t,r]*(1-s.ch4red[t-1,r]);        
  end;

  // Calculate n2o emissions
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    s.n2o[t,r]:=s.n2oem[t,r]*(1-s.n2ored[t-1,r]);        
  end;

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin  
    if s.emission[t,r]/s.income[t,r] - s.minint[t-1] <= 0 then
      s.taxpar[t,r]:=0.784
    else
      s.taxpar[t,r]:=0.784 - 0.084*Math.Sqrt(s.emission[t,r]/s.income[t,r] - s.minint[t-1]);
  end;

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin  
    s.co2red[t,r]:=s.currtax[t,r]*s.emission[t,r]*s.know[t-1,r]*s.globknow[t-1]/2/s.taxpar[t,r]/s.income[t,r]/1000;

    if s.co2red[t,r] < 0 then
      s.co2red[t,r]:= 0
    else if s.co2red[t,r] > 0.99 then
      s.co2red[t,r]:= 0.99;
  end;

  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin  
    s.ryg[t,r]:=s.taxpar[t,r]*Math.Pow(s.co2red[t,r],2)/s.know[t-1,r]/s.globknow[t-1];
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    s.perm[t,r]:= 1-0.01*s.currtax[t,r]/(1+0.01*s.currtax[t,r]);
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    s.reei[t,r]:= s.perm[t,r]*0.5*s.co2red[t,r];
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    if s.currtax[t,r] < 100 then
      s.rcei[t,r]:=s.perm[t,r]*0.5*Math.Pow(s.co2red[t,r],2)
    else
      s.rcei[t,r]:=s.perm[t,r]*0.5*s.co2red[t,r];
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    s.seei[t,r]:=0.9*s.seei[t-1,r] + (1-s.perm[t,r])*0.5*s.co2red[t,r]*1.7;
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    if s.currtax[t,r] < 100 then
      s.scei[t,r]:= 0.9*s.scei[t-1,r] + (1-s.perm[t,r])*0.5*Math.Pow(s.co2red[t,r],2)
    else
      s.scei[t,r]:= 0.9*s.scei[t-1,r] + (1-s.perm[t,r])*0.5*s.co2red[t,r]*1.7;
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin   
    s.know[t,r]:=s.know[t-1,r]*Math.Sqrt(1+s.knowpar*s.co2red[t,r]);
    
    if s.know[t,r] > 3.16 then
      s.know[t,r]:=3.16;
  end;
      
  s.globknow[t]:=s.globknow[t-1];
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin            
    s.globknow[t]:=s.globknow[t]*Math.Sqrt(1+s.knowgpar*s.co2red[t,r]);
  end;
  if s.globknow[t] > 3.16 then
    s.globknow[t]:=3.16;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin             
    s.ch4red[t,r]:= s.gwpch4*s.currtax[t,r]*s.ch4em[t,r]/2/s.ch4par1[r]/s.ch4par2[r]/s.ch4par2[r]/s.income[t,r]/1000;

    if s.ch4red[t,r] < 0 then
      s.ch4red[t,r]:= 0
    else if s.ch4red[t,r] > 0.99 then
      s.ch4red[t,r]:= 0.99;
  end;
    
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin             
    s.n2ored[t,r]:= s.gwpn2o*s.currtax[t,r]*s.n2oem[t,r]/2/s.n2opar1[r]/s.n2opar2[r]/s.n2opar2[r]/s.income[t,r]/1000;
    
    if s.n2ored[t,r] < 0 then
      s.n2ored[t,r]:= 0
    else if s.n2ored[t,r] > 0.99 then
      s.n2ored[t,r]:= 0.99;
  end;
         
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin             
    s.ch4cost[t,r]:=s.ch4par1[r]*Math.Pow(s.ch4par2[r],2)*Math.Pow(s.ch4red[t,r],2);
  end;
  
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin              
    s.n2ocost[t,r]:=s.n2opar1[r]*Math.Pow(s.n2opar2[r],2)*Math.Pow(s.n2ored[t,r],2);
  end;
  
  var minint:=Double.PositiveInfinity;    
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin              
    if s.emission[t,r]/s.income[t,r] < minint then
      minint:=s.emission[t,r]/s.income[t,r];
  end;
  s.minint[t]:=minint;
end;

end.