// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

uses
  System.Collections.Generic,
  System.Configuration,
  System.IO;

type
  
  MarginalGas = public(C, CH4, N2O, SF6);
  
  TMarginalRun = class
  private
    { Private Declarations }
    m_YearRegionCsv: TextWriter;
    m_YearRegionWeightingschemeCsv: TextWriter;
    m_YearRegionSectorWeightingSchemeCsv: TextWriter;
    m_AggregateDamageCsv: TextWriter;
    m_GlobalInputCsv: TextWriter;
    m_RegionInputCsv: TextWriter;
    m_YearInputCsv: TextWriter;
    m_RegionYearInputCsv: TextWriter;    
    
    _run: Run;
    _gas: MarginalGas;
    _emissionyear: Integer;
    _outputPath: String;
    _parameters: FundParameter;
    
    procedure MargRun(Output: ModelOutput; addmp: boolean);
    method DoOneRun(RunId: Integer; i_weightingCombinations: array of WeightingCombination): array of Double;

    procedure WriteAggregateDamage(RunId: Integer; WeightingschemeId: Integer; Damage: Double; WeightingCombinations: array of WeightingCombination);
    procedure WriteMarginalDamage(RunId: Integer; i_Damage: Damage; WeightingschemeId: Integer; Weight: Double; WeightingCombinations: array of WeightingCombination);
    procedure WriteYearRegionWeightingschemeData(RunId: Integer; RegionId: Integer; WeightingSchemeId: Integer; i_weightingCombinations: array of WeightingCombination; Year: Integer);
    procedure WriteYearRegionData(RunId: Integer; Year: Integer; Region: Integer; Income: Double; Population: Double);

  public
    constructor(runData: Run; gas: MarginalGas; emissionyear: Integer; outputPath: string; parameters: FundParameter);

    method Run: array of double;

    property AggregateDamageCsv: Textwriter read m_AggregateDamageCsv write m_AggregateDamageCsv;
    property YearRegionCsv: TextWriter read m_YearRegionCsv write m_YearRegionCsv;
    property YearRegionWeightingSchemeCsv: TextWriter read m_YearRegionWeightingschemeCsv write m_YearRegionWeightingschemeCsv;
    property YearRegionSectorWeightingSchemeCsv: TextWriter read m_YearRegionSectorWeightingSchemeCsv write m_YearRegionSectorWeightingSchemeCsv;
    property GlobalInputCsv: TextWriter read m_GlobalInputCsv write m_GlobalInputCsv;
    property RegionInputCsv: TextWriter read m_RegionInputCsv write m_RegionInputCsv;
    property YearInputCsv: TextWriter read m_YearInputCsv write m_YearInputCsv;
    property RegionYearInputCsv: TextWriter read m_RegionYearInputCsv write m_RegionYearInputCsv;    
  end;

implementation

constructor TMarginalRun(runData: Run; gas: MarginalGas; emissionyear: Integer; outputPath: string; parameters: FundParameter);
begin
  inherited constructor();

  _parameters:=parameters;
  _run:=runData;
  _gas:=gas;
  _emissionyear:=emissionyear;
  _outputPath:=outputPath;
end;

// DA: Run the model once, write away income (GDP), population, market impacts, non market
// impacts (prn files)
// Run the model again, with a slightly higher emissions, additional emissions are in the file
// addco2.001 (million ton of carbons) Write the results of the second run out to (out files)
method TMarginalRun.Run: Array of Double;
var
  i: Integer;
  i_weightingCombinations: array of WeightingCombination;
  lDimWeightingschemeCsv: TextWriter;
begin
  _parameters.SetToBestGuessValue;
  
  Fund28LegacyWeightingCombinations.GetWeightingCombinationsFromName(_run.WeightingCombination, out i_weightingCombinations, _parameters);

  lDimWeightingschemeCsv:=new StreamWriter(Path.Combine(_outputPath, 'Output - Dim Weightingscheme.csv'));
  for i:=0 to High(i_weightingCombinations) do
  begin
    lDimWeightingschemeCsv.WriteLine(i.ToString + ';' + i_weightingCombinations[i].Name + '');
  end;
  lDimWeightingschemeCsv.Close;

  var meanResults: Array of Double:=new double[i_weightingCombinations.Length];
  for l: integer:=0 to High(i_weightingCombinations) do
  begin
    meanResults[l]:=0.0;
  end;  
  
  var totalToTrim: Integer:=20;
  var meanCount: Integer:=0;
  var trimInitialList:=new List<array of Double>;
  var trimTopResults:=new List<array of Double>;
  var trimBottomResults:=new List<array of Double>;
  
  for i:=0 TO _run.MonteCarloRuns DO
  begin
    if i=0 then
    begin
      Console.WriteLine('Best guess run ');

      _parameters.SetToBestGuessValue;
      _parameters.WriteValuesToCsv(_run.Scenario.Id, i,
           m_GlobalInputCsv, m_RegionInputCsv, m_YearInputCsv, m_RegionYearInputCsv);      
    end
    else
    begin
      try
        Console.SetCursorPosition(0, Console.CursorTop - 1);
      except
        on E: IOException do
        begin          
        end;
      end;
      Console.WriteLine('Run {0}                ', i.ToString);

      _parameters.SetToRandomValue;
      _parameters.WriteValuesToCsv(_run.Scenario.Id, i,
           m_GlobalInputCsv, m_RegionInputCsv, m_YearInputCsv, m_RegionYearInputCsv);      
    end;
   
    var res:=DoOneRun(i, i_weightingCombinations);
    
    Result:=res;
    
    if not double.IsNaN(res[0]) and not double.IsInfinity(res[0]) then
    begin
      if (trimBottomResults.Count + trimTopResults.Count)=0 then
      begin
        trimInitialList.Add(res);
        
        if trimInitialList.Count=totalToTrim then
        begin
          trimInitialList.Sort(
            method(r1,r2: Array of Double): Integer;
            begin            
              Result:=iif(r1[0]<r2[0], -1, 1);
            end);
            
          trimBottomResults.AddRange(trimInitialList.GetRange(0,totalToTrim/2));
          trimTopResults.AddRange(trimInitialList.GetRange(totalToTrim/2,totalToTrim/2));
          
          trimBottomResults.Sort(
            method(r1,r2: Array of Double): Integer;
            begin            
              Result:=iif(r1[0]>r2[0], -1, 1);
            end);
          
          trimInitialList:=nil;
        end;
        
      end
      else
      begin
        // is the result smaller than the largest bottom result?
        if res[0]<trimBottomResults[0][0] then
        begin
          var tres:=trimBottomResults[0];
          trimBottomResults[0]:=res;
          res:=tres;    
          
          trimBottomResults.Sort(
            method(r1,r2: Array of Double): Integer;
            begin            
              Result:=iif(r1[0]>r2[0], -1, 1);
            end);              
        end
        else if res[0]>trimTopResults[0][0] then
        begin
          var tres:=trimTopResults[0];
          trimTopResults[0]:=res;
          res:=tres;  
          
          trimTopResults.Sort(
            method(r1,r2: Array of Double): Integer;
            begin            
              Result:=iif(r1[0]<r2[0], -1, 1);
            end);                
        end;

        inc(meanCount);
        for l: integer:=0 to High(i_weightingCombinations) do
        begin
          meanResults[l]:=meanResults[l] + res[l];
        end;    
      end;
    end;
    
  end;
  
  for l: integer:=0 to High(i_weightingCombinations) do
  begin
    meanResults[l]:=meanResults[l]/double(meanCount);
  
    WriteAggregateDamage(-1,l,meanResults[l], i_weightingCombinations);
  end;
  
  if _run.MonteCarloRuns>0 then
    Result:=meanResults;
end;

method TMarginalRun.DoOneRun(RunId: Integer; i_weightingCombinations: array of WeightingCombination): Array of Double;
var
  i_output2: ModelOutput;
  i: Integer;
  i_Damage: Damage;
  i_marginalDamages: Damages;
  i_aggregatedDamage: double;
  l: Integer;
  k: Integer;
  m: Integer;
  i_output1: ModelOutput;
begin
  // Create Output object for run 1, set addmp to 0 so that
  // the extra greenhouse gases are not emitted and then run
  // the model
  i_output1 := new ModelOutput;
  MargRun(i_output1, false);
  // Create Output object for run 2, set addmp to 1 so that
  // the extra greenhouse gases for the marginal run are
  // emitted and then run the model
  i_output2 := new ModelOutput;
  
  _parameters.emissionperiod:=_emissionyear;
  case _gas of
    MarginalGas.C: _parameters.marginalgas:=0;
    MarginalGas.CH4: _parameters.marginalgas:=1;
    MarginalGas.N2O: _parameters.marginalgas:=2;
    MarginalGas.SF6: _parameters.marginalgas:=3;
  end;
  
  MargRun(i_output2, true);
  
  Fund28LegacyWeightingCombinations.GetWeightingCombinationsFromName(_run.WeightingCombination, out i_weightingCombinations, _parameters); 

  // Take out growth effect effect of run 2 by transforming
  // the damage from run 2 into % of GDP of run 2, and then
  // multiplying that with GDP of run 1
  for i := 0 to i_output2.Damages.Count - 1 do
  begin
    i_Damage := i_output2.Damages.Item[i];
    i_Damage.Damage := (i_Damage.Damage / i_output2.Incomes.Item[i_Damage.Year, i_Damage.Region]) * i_output1.Incomes.Item[i_Damage.Year, i_Damage.Region];
  end;
  
  // Calculate the marginal damage between run 1 and 2 for each
  // year/region/sector
  i_marginalDamages := Damages.CalculateMarginalDamage(i_output1.Damages, i_output2.Damages);
  
  var i_weightedAggregatedDamages: Array of Double:=new double[i_weightingCombinations.Length];  
  
  for i := 0 to High(i_weightingCombinations) do
  begin
    i_weightingCombinations[i].CalculateWeights(i_output1);
    i_aggregatedDamage := i_weightingCombinations[i].AddDamagesUp(i_marginalDamages, _run.YearsToAggregate, Convert.ToInt32(_parameters.emissionPeriod));

    i_weightedAggregatedDamages[i]:=i_aggregatedDamage;
    
    WriteAggregateDamage(RunId, i, i_aggregatedDamage, i_weightingCombinations);

    // Console.Write(i_weightingCombinations[i].Name + ': ');
    // Console.WriteLine(Convert.ToString(i_aggregatedDamage));
  end;
  for l := 0 to i_marginalDamages.Count - 1 do
  begin
    i_Damage := i_marginalDamages.Item[l];
    if (i_Damage.Year >= _parameters.emissionPeriod-1950) AND (i_Damage.Year < _parameters.emissionPeriod-1950 + _run.YearsToAggregate) then
    begin
      for k := 0 to High(i_weightingCombinations) do
        WriteMarginalDamage(RunId, i_Damage, k, i_weightingCombinations[k].Weight[i_Damage.Year, i_Damage.Region], i_weightingCombinations );
    end;
  end;
  for l := Convert.ToInt32(_parameters.emissionPeriod-1950) to Convert.ToInt32(_parameters.emissionPeriod-1950)-1950 + _run.YearsToAggregate do
  begin
    for k := 1 to LegacyConstants.NoReg do
    begin
      WriteYearRegionData(RunId, l, k, i_output1.Incomes.Item[l, k], i_output1.Populations.Item[l, k]);
      for m := 0 to High(i_weightingCombinations) do
        WriteYearRegionWeightingschemeData(RunId, k, m, i_weightingCombinations, l);
    end;
  end;
  Result:=i_weightedAggregatedDamages;
end;

procedure TMarginalRun.WriteYearRegionData(RunId: Integer; Year: Integer; Region: Integer; Income: Double; Population: Double);
begin
  if m_YearRegionCsv<>nil then
  begin
    m_YearRegionCsv.Write(_run.Scenario.Id);
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write(Convert.ToInt32(_parameters.marginalGas));
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write(Convert.ToInt32(_parameters.emissionPeriod));
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write(RunId);
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write((Year + 1950));
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write(Region);
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write('{0:f15}', Object(Income));
    m_YearRegionCsv.Write(';');
    m_YearRegionCsv.Write('{0:f15}', Object(Population));
    m_YearRegionCsv.WriteLine;
  end;
end;

procedure TMarginalRun.WriteYearRegionWeightingschemeData(RunId: Integer; RegionId: Integer; WeightingSchemeId: Integer; i_weightingCombinations: array of WeightingCombination; Year: Integer);
var
  i_e: double;
begin
  if m_YearRegionWeightingschemeCsv<>nil then
  begin
    m_YearRegionWeightingschemeCsv.Write(_run.Scenario.Id);
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write(Convert.ToInt32(_parameters.marginalGas));
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write(Convert.ToInt32(_parameters.emissionPeriod));
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write(RunId);
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write((Year + 1950));
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write(RegionId);
    m_YearRegionWeightingschemeCsv.Write(';');
    m_YearRegionWeightingschemeCsv.Write(WeightingSchemeId);
    m_YearRegionWeightingschemeCsv.Write(';');
    i_e := i_weightingCombinations[WeightingSchemeId].Weight[Year, RegionId];
    m_YearRegionWeightingschemeCsv.Write('{0:f15}', Object(i_e));
    m_YearRegionWeightingschemeCsv.WriteLine;
  end;
end;

procedure TMarginalRun.WriteMarginalDamage(RunId: Integer; i_Damage: Damage; WeightingschemeId: Integer; Weight: Double; WeightingCombinations: array of WeightingCombination);
var
  i_e: double;
begin
  if m_YearRegionSectorWeightingSchemeCsv<>nil then
  begin
    m_YearRegionSectorWeightingSchemeCsv.Write(_run.Scenario.Id);
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    
    if _run.OutputVerbal then
    begin
      if RunId=0 then
        m_YearRegionSectorWeightingSchemeCsv.Write('"Best guess"')
      else
        m_YearRegionSectorWeightingSchemeCsv.Write(RunId);
    end
    else
      m_YearRegionSectorWeightingSchemeCsv.Write(RunId);
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    
    
    if _run.OutputVerbal then
    begin
      case Convert.ToInt32(_parameters.marginalGas) of
        0: m_YearRegionSectorWeightingSchemeCsv.Write('"C"');
        else
          raise new NotImplementedException('Only carbon supported for output at this point');
      end;
    end
    else
      m_YearRegionSectorWeightingSchemeCsv.Write(Convert.ToInt32(_parameters.marginalGas));
    
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    
    m_YearRegionSectorWeightingSchemeCsv.Write(Convert.ToInt32(_parameters.emissionPeriod));
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    
    m_YearRegionSectorWeightingSchemeCsv.Write((i_Damage.Year + 1950));
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    if _run.OutputVerbal then
    begin
          m_YearRegionSectorWeightingSchemeCsv.Write(LegacyRegionNames.Name[i_Damage.Region]);
    end
    else
      m_YearRegionSectorWeightingSchemeCsv.Write(i_Damage.Region);
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    if _run.OutputVerbal then
    begin
      m_YearRegionSectorWeightingSchemeCsv.Write(Enum.GetName(typeof(Sector), i_Damage.Sector));
    end
    else
      m_YearRegionSectorWeightingSchemeCsv.Write(Ord(i_Damage.Sector));
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    if _run.OutputVerbal then
    begin
      m_YearRegionSectorWeightingSchemeCsv.Write('"' + WeightingCombinations[WeightingschemeId].Name + '"');
    end
    else
      m_YearRegionSectorWeightingSchemeCsv.Write(WeightingschemeId);
    m_YearRegionSectorWeightingSchemeCsv.Write(';');
    i_e := i_Damage.Damage * Weight;
    m_YearRegionSectorWeightingSchemeCsv.Write('{0:f15}', Object(i_e));
    m_YearRegionSectorWeightingSchemeCsv.WriteLine;
  end;
end;

// Output:  MargRun1 writes a number of model results into the object referenced
//          by that variable.
procedure TMarginalRun.MargRun(Output: ModelOutput; addmp: boolean);
var
  fundWorkflow: FundWorkflow;
begin
  fundWorkflow:=new FundWorkflow(_parameters);
  
  fundWorkflow.ModelFields.SetNonDimensionalFieldValue('climate','addmp', addmp);
  
  fundWorkflow.Run(Output);         
end;

procedure TMarginalRun.WriteAggregateDamage(RunId: Integer; WeightingschemeId: Integer;
  Damage: Double; WeightingCombinations: array of WeightingCombination);
begin
  if not _run.CalculateMeanForMonteCarlo and (RunId=-1) then
    Exit
  else if not _run.OutputAllMonteCarloRuns and (RunId>0) then
    Exit;
    
  if m_AggregateDamageCsv<>nil then
  begin
    if _run.OutputVerbal then
      m_AggregateDamageCsv.Write(_run.Scenario.Name)
    else
      m_AggregateDamageCsv.Write(_run.Scenario.Id);
      
    m_AggregateDamageCsv.Write(';');

    if _run.OutputVerbal then
    begin
      case Convert.ToInt32(_parameters.marginalGas) of
        0: m_AggregateDamageCsv.Write('"C"');
        1: m_AggregateDamageCsv.Write('"CH4"');
        2: m_AggregateDamageCsv.Write('"N2O"');
        3: m_AggregateDamageCsv.Write('"SF6"');
        else
          raise new NotImplementedException('Only carbon supported for output at this point');
      end;
    end
    else
      m_AggregateDamageCsv.Write(Convert.ToInt32(_parameters.marginalGas));

    m_AggregateDamageCsv.Write(';');
    m_AggregateDamageCsv.Write(Convert.ToInt32(_parameters.emissionPeriod));
    m_AggregateDamageCsv.Write(';');

    if _run.OutputVerbal then
    begin
      if RunId=0 then
        m_AggregateDamageCsv.Write('"Best guess"')
      else if RunId=-1 then
        m_AggregateDamageCsv.Write('"Mean"')
      else
        m_AggregateDamageCsv.Write(RunId);
    end
    else
      m_AggregateDamageCsv.Write(RunId);

    m_AggregateDamageCsv.Write(';');

    if _run.OutputVerbal then
    begin
      m_AggregateDamageCsv.Write('"' + WeightingCombinations[WeightingschemeId].Name + '"');
    end
    else
      m_AggregateDamageCsv.Write(WeightingschemeId);

    m_AggregateDamageCsv.Write(';');
    m_AggregateDamageCsv.Write('{0:f15}', Object(Damage));
    m_AggregateDamageCsv.WriteLine;

  end;
end;

end.
