// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

uses
  System.Collections.Generic,
  Esmf,
  Fund.CommonDimensions;

type
  MarginalDamage2 = public class
  private
    _parameters: FundParameter;
    _run: Run;
    _gas: MarginalGas;
    _emissionyear: Timestep;
    _useNegishiWeights: Boolean;
    _prtp: Double := 0.001;
    _eta: Double := 1.0;
  protected
  public
    constructor;
    method Start: IDictionary<Region,double>;
    
    property Parameters: FundParameter read _parameters write _parameters;
    property Run: Run read _run write _run;
    property Gas: MarginalGas read _gas write _gas;
    property EmissionYear: Timestep read _emissionyear write _emissionyear;
    property UseNegishiWeights: Boolean read _useNegishiWeights write _useNegishiWeights;
    property AdditionalInitMethod: Action<FundWorkflow> read write;
    property Prtp: Double read _prtp write _prtp;
    property Eta: Double read _eta write _eta;
  end;
  
implementation

constructor MarginalDamage2;
begin
end;

method MarginalDamage2.Start: IDictionary<Region,double>; 
begin

  _parameters.SetToBestGuessValue;
  
  var setutilityparameters:=method(run: FundWorkflow; utilitycalibrationmultiplicative: Double);
    begin
      run.ModelFields.SetNonDimensionalFieldValue('globalwelfare','prtp', prtp);
      run.ModelFields.SetNonDimensionalFieldValue('globalwelfare','elasticityofmarginalutility', eta);
      run.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','prtp', prtp);
      run.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','elasticityofmarginalutility', eta); 
      run.ModelFields.SetNonDimensionalFieldValue('globalwelfare','utilitycalibrationmultiplicative',utilitycalibrationmultiplicative);            
      run.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','utilitycalibrationmultiplicative',utilitycalibrationmultiplicative);            
      run.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','starttimestep',_emissionyear);
      run.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutpopulationperturbation', true);
      run.ModelFields.SetNonDimensionalFieldValue('socioeconomic','consleak', 0.0);      
    end;

//  var calibrationrunone:=new FundWorkflow(_parameters);
//  setutilityparameters(calibrationrunone, prtp, eta, 1.0);
//  calibrationrunone.Run(nil);
//  var multicali:=1.0/calibrationrunone.RegionalWelfareState.marginalwelfare[_emissionyear, Region.USA];

  var multicali:=1.0;

  var run1:=new FundWorkflow(_parameters);
  setutilityparameters(run1, multicali);
  
  var run2:=new FundWorkflow(_parameters);  
  setutilityparameters(run2, multicali);
  run2.ModelFields.SetNonDimensionalFieldValue('climate','addmp', true);
  run2.ModelFields.Set1DimensionalParameter<Timestep,double>('climate', 'addco2',
    method(t: Timestep): Double;
    begin
      if (_gas=MarginalGas.C) and (_emissionyear<=t) and (t<_emissionyear+10) then
        Result:=1.0
      else
        Result:=0.0;    
    end);
  
  if _useNegishiWeights then
  begin
    var calibrationrun:=new FundWorkflow(_parameters);
    setutilityparameters(calibrationrun, multicali);
    if AdditionalInitMethod<>nil then
      AdditionalInitMethod(calibrationrun);
    calibrationrun.Run(nil);    
    run1.ModelFields.Set1DimensionalParameter<Region,double>('regionalwelfare','welfareweight',
      method(r: Region): Double;
      begin
        Result:=calibrationrun.RegionalWelfareState.marginalwelfare[_emissionyear,Region.USA]/calibrationrun.RegionalWelfareState.marginalwelfare[_emissionyear,r];  
      end);
    run2.ModelFields.Set1DimensionalParameter<Region,double>('regionalwelfare','welfareweight',
      method(r: Region): Double;
      begin
        Result:=calibrationrun.RegionalWelfareState.marginalwelfare[_emissionyear,Region.USA]/calibrationrun.RegionalWelfareState.marginalwelfare[_emissionyear,r];  
      end);  
  end;

  if AdditionalInitMethod<>nil then
  begin
    AdditionalInitMethod(run1);
    AdditionalInitMethod(run2);
  end;
  
  run1.Run(nil);  
  run2.Run(nil);
  
  var sccInUtils:=(run1.RegionalWelfareState.totalwelfare-run2.RegionalWelfareState.totalwelfare)/10000000.0;
  
  var res:=new Dictionary<Region,double>;
  for each r: Region in Enum.GetValues(typeof(Region)) do
  begin
    res.Add(r,sccInUtils/run2.RegionalWelfareState.marginalwelfare[_emissionyear, r]);
  end;  
  
  Result:=res;
end;    

end.