// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

uses System.Collections,
  System.Collections.Generic,
  Esmf,
  System.IO;

type

  ParameterAttribute = class( Attribute )
  private
    m_Name: string;
  public
    constructor;
    property Name: string read m_Name write m_Name;
  end;

  RegionParameter = public class
  private
    m_values: array [0..LegacyConstants.NoReg-1] of double;
  protected
    procedure SetItem(Region: integer; Value: double);
    function GetItem(Region: integer): double;
  public
    constructor;

    property Item[Region: integer]: double read GetItem write SetItem; default;
  end;
  
  RegionRegionParameter = public class
  private
    m_values: array [0..LegacyConstants.NoReg-1, 0..LegacyConstants.NoReg-1] of double;
  protected
    procedure SetItem(region1: Integer; region2: integer; value: double);
    function GetItem(region1: integer; region2: integer): double;
  public
    constructor;
    
    property Item[region1: integer; region2: integer]: double read GetItem write SetItem; default;
  end;

  YearParameter = public class
  private
    m_values: array [0..LegacyConstants.NYear] of double;
  protected
    procedure SetItem(Year: integer; Value: double);
    function GetItem(Year: integer): double;
  public
    constructor;
    
    property Item[Year: integer]: double read GetItem write SetItem; default;
  end;

  YearRegionParameter = public class
  private
    m_values: array [0..LegacyConstants.NYear, 0..LegacyConstants.NoReg-1] of double;
  protected
    procedure SetItem(Year, Region: integer; Value: double);
    function GetItem(Year, Region: integer): double;
  public
    constructor;

    property Item[Year, Region: integer]: double read GetItem write SetItem; default;
  end;
  
  ParameterIdAndName = public class
  public
    property Id: Integer;
    property Name: String;
  end;
  
  ParameterKey = public abstract class 
      private
        m_NameAndId: ParameterIdAndName;
      protected
        method GetName: String;
        method GetId: Integer;
      public
        constructor(NameAndId: ParameterIdAndName);

        property Name: string read GetName;
        property Id: Integer read GetId;
      end;

      GlobalParameterKey = public class( ParameterKey )
      public
        constructor(NameAndId: ParameterIdAndName);
      end;

      RegionParameterKey = public class( ParameterKey )
      private
        m_regionId: Integer;
        m_region: string;
      public
        constructor(Region: string; NameAndId: ParameterIdAndName);

        property RegionName: string read m_region;
        property RegionId: Integer read m_regionId;
      end;
      
      RegionRegionParameterKey = public class(ParameterKey)
      private
        m_region1: string;
        m_region1Id: integer;
        m_region2: string;
        m_region2Id: Integer;
      public
        constructor(region1Data, region2Data: string; NameAndId: ParameterIdAndName);
        
        property Region1: string read m_region1;
        property Region1Id: integer read m_region1Id;
        property Region2: string read m_region2;
        property Region2Id: integer read m_region2Id;
      end;

      YearParameterKey = public class( ParameterKey )
      private
        m_year: integer;
      public
        constructor(YearData: integer; NameAndId: ParameterIdAndName);

        property Year: integer read m_year;
      end;

      YearRegionParameterKey = public class( ParameterKey )
      private
        m_year: integer;
        m_region: string;
        m_regionId: Integer;
      public
        constructor(YearData: integer; RegionData: string; NameAndId: ParameterIdAndName);

        property Year: integer read m_year;
        property Region: string read m_region;
        property RegionId: Integer read m_RegionId;
      end;

      Parameter = public abstract class 
      protected
        m_key: ParameterKey;
      protected
        _rand: Random;
        function GetValue: double; virtual; abstract;
      public
        constructor(KeyData: ParameterKey; rand: Random);

        procedure SetToBestGuessValue; virtual; abstract;
        procedure SetToRandomValue; virtual; abstract;

        property Key: ParameterKey read m_key;
        property Value: double read GetValue;
      end;

      ConstantParameter = public class( Parameter )
      private
        m_value: double;
      protected
        function GetValue: double; override;
      public
        constructor(Key: ParameterKey; Value: double);

        procedure SetToBestGuessValue; override;
        procedure SetToRandomValue; override;
      end;

      NormalDistributionParameter = public class( Parameter )
      private
        m_value: double;
        m_bestGuess: double;
        m_standardDeviation: double;
        m_useLowerBound, m_useUpperBound: Boolean;
        m_lowerBound, m_upperBound: double;
      protected
        function GetValue: double; override;
      public
        constructor(Key: ParameterKey; Arguments: array of string; rand: Random);
        constructor(Key: ParameterKey; BestGuess, StandardDeviation: double; rand: Random);
        constructor(Key: ParameterKey; BestGuess, StandardDeviation: double; UseLowerBound: Boolean; LowerBound: double; UseUpperBound: Boolean; UpperBound: double; rand: Random);

        procedure SetToBestGuessValue; override;
        procedure SetToRandomValue; override;
      end;

      ExponentialDistributionParameter = public class( Parameter )
      private
        m_value: double;
        m_lambda: double;
        m_useLowerBound, m_useUpperBound: Boolean;
        m_lowerBound, m_upperBound: double;
      protected
        function GetValue: double; override;
      public
        constructor(Key: ParameterKey; Arguments: array of string; rand: Random);

        procedure SetToBestGuessValue; override;
        procedure SetToRandomValue; override;
      end;

      GammaDistributionParameter = public class( Parameter )
      private
        m_value: double;
        m_alpha, m_beta: double;
      protected
        function GetValue: double; override;
      public
        constructor(Key: ParameterKey; Arguments: array of string; rand: Random);

        procedure SetToBestGuessValue; override;
        procedure SetToRandomValue; override;
      end;

      TriangularDistributionParameter = public class( Parameter )
      private
        m_value: double;
        m_bestGuess: double;
        m_min, m_max: double;
      protected
        function GetValue: double; override;
      public
        constructor(Key: ParameterKey; Arguments: array of string; rand: Random);

        procedure SetToBestGuessValue; override;
        procedure SetToRandomValue; override;
      end;  

  Parameters = public class
  private
    m_numberOfRegions: Integer;
    m_numberOfYears: Integer;
    m_globalParameters: Hashtable;
    m_regionParameters: Hashtable;
    m_regionRegionParameters: Dictionary<string, Parameter>;
    m_yearParameters: Hashtable;
    m_yearRegionParameters: Hashtable;
    m_allParameterNameAndId: List<ParameterIdAndName>;
    m_allParameterIdsByName: Dictionary<string, integer>:=new Dictionary<string, integer>;
  protected
    function GetItem(Name: string): double;
    method GetItemByRegionRegion(region1, region2, name: string): double;

    procedure Add(Parameter: Parameter); 
    procedure Add(ParameterKey: ParameterKey; ValueString: string ); 

    procedure UseReflectionToSetFieldsCurrentValues;
    procedure UseReflectionToCreateObjects;
    
    method ReadExcel2003File(Filename: string);
    method ReadExcel2007File(Filename: string);
  public
    _rand: Random;
    
    const c_normalDistributionName: string = 'NormalDistribution';
    const c_triangularDistributionName: string = 'TriangularDistribution';
    const c_exponentialDistributionName: string = 'ExponentialDistribution';
    const c_gammaDistributionName: string = 'GammaDistribution';
    
    
    constructor(NumberOfYears, NumberOfRegions: Integer; rand: Random);

    procedure SetToBestGuessValue;
    procedure SetToRandomValue;

    procedure ReadExcelFile(FileName: string);
    procedure WriteCurrentValuesToFile(FileName: string);
    procedure WriteValuesToCsv (ScenarioId: Integer; RunId: Integer;
                                GlobalCsv, RegionCsv , YearCsv, RegionYearCsv: TextWriter);
    
    function GetItemByYearAndRegion(Year: integer; Region, Name: string): double;
    function GetItemByYear(Year: integer; Name: string): double;
    function GetItemByRegion(Region, Name: string): double;
    
    property ParameterIdAndNames: ICollection<ParameterIdAndName> read m_allParameterNameAndId;
  end;

implementation

uses
    System.Globalization,
    System.Data.OleDb,
    System.Data;


{ Parameters }

procedure Parameters.Add(Parameter: Parameter);
var
  i_globalParameterkey: GlobalParameterKey;
  i_regionParameterKey: RegionParameterKey;
  i_yearParameterKey: YearParameterKey;
  i_yearRegionParameterKey: YearRegionParameterKey;
begin
  if Parameter.Key is GlobalParameterKey then
  begin
    i_globalParameterkey:=Parameter.Key as GlobalParameterKey;
    
    if m_globalParameters.ContainsKey(i_globalParameterkey.Name) then
       m_globalParameters.Item[i_globalParameterkey.Name]:= Parameter
    else
      m_globalParameters.Add(i_globalParameterkey.Name, Parameter);
  end
  else if Parameter.Key is RegionParameterKey then
  begin
    i_regionParameterKey:=Parameter.Key as RegionParameterKey;
    var i_regionParameterKeyAsString:=i_regionParameterKey.RegionName + ';' + i_regionParameterKey.Name;
    
    if m_regionParameters.ContainsKey(i_regionParameterKeyAsString) then
      m_regionParameters.Item[i_regionParameterKeyAsString]:=Parameter
    else
      m_regionParameters.Add(i_regionParameterKeyAsString, Parameter);
  end
  else if Parameter.Key is RegionRegionParameterKey then
  begin
    var i_parameterKey:=Parameter.Key as RegionRegionParameterKey;
    var i_parameterKeyAsString:=String.Format('{0};{1};{2}', i_parameterKey.Region1, i_parameterKey.Region2, i_parameterKey.Name);
    
    if m_regionRegionParameters.ContainsKey(i_parameterKeyAsString) then
      m_regionRegionParameters.Item[i_parameterKeyAsString]:=Parameter
    else
      m_regionregionParameters.Add(i_parameterKeyAsString, Parameter);
  end
  else if Parameter.Key is YearParameterKey then
  begin
    i_yearParameterKey:=Parameter.Key as YearParameterKey;
    var i_yearParameterKeyAsString:=Convert.ToString(i_yearParameterKey.Year) + ';' + i_yearParameterKey.Name;
    
    if m_yearParameters.ContainsKey(i_yearParameterKeyAsString) then
      m_yearParameters.Item[i_yearParameterKeyAsString]:=Parameter
    else
      m_yearParameters.Add(i_yearParameterKeyAsString, Parameter);
  end
  else if Parameter.Key is YearRegionParameterKey then
  begin
    i_yearRegionParameterKey:=Parameter.Key as YearRegionParameterKey;
    var i_yearRegionParameterKeyAsString:=Convert.ToString(i_yearRegionParameterKey.Year) + ';' + i_yearRegionParameterKey.Region + ';' + i_yearRegionParameterKey.Name;
    
    if m_yearRegionParameters.ContainsKey(i_yearRegionParameterKeyAsString) then
      m_yearRegionParameters.Item[i_yearRegionParameterKeyAsString]:=Parameter
    else
      m_yearRegionParameters.Add(i_yearRegionParameterKeyAsString, Parameter);
  end
  else
  begin
  end;
end;

procedure Parameters.Add(ParameterKey: ParameterKey; ValueString: string);
var
  i_value: string;
  i_fake: Double;
begin
  i_value:=ValueString.Trim();
  
  // TODO Control, whether NumberStyles.Float is appropriate
  if Not (Double.TryParse(i_value, NumberStyles.Float, CultureInfo.InstalledUICulture, out i_fake) or (Double.TryParse(i_value, NumberStyles.Float,CultureInfo.InvariantCulture, out i_fake))) then
  begin
    if i_value.StartsWith(c_normalDistributionName + '(') and i_value.EndsWith(')') then
    begin
      i_value:=i_value.Substring( c_normalDistributionName.Length + 1, i_value.Length - (c_normalDistributionName.Length + 2));
      Add(new NormalDistributionParameter(ParameterKey, i_value.Split([';']), _rand));
    end
    else if i_value.StartsWith(c_triangularDistributionName + '(') and i_value.EndsWith(')') then
    begin
      i_value:=i_value.Substring( c_triangularDistributionName.Length + 1, i_value.Length - (c_triangularDistributionName.Length + 2));
      Add(new TriangularDistributionParameter(ParameterKey, i_value.Split([';']), _rand));
    end
    else if i_value.StartsWith(c_exponentialDistributionName + '(') and i_value.EndsWith(')') then
    begin
      i_value:=i_value.Substring( c_exponentialDistributionName.Length + 1, i_value.Length - (c_exponentialDistributionName.Length + 2));
      Add(new ExponentialDistributionParameter(ParameterKey, i_value.Split([';']), _rand));
    end
    else if i_value.StartsWith(c_GammaDistributionName + '(') and i_value.EndsWith(')') then
    begin
      i_value:=i_value.Substring( c_GammaDistributionName.Length + 1, i_value.Length - (c_GammaDistributionName.Length + 2));
      Add(new GammaDistributionParameter(ParameterKey, i_value.Split([';']), _rand));
    end
    else
      raise new ApplicationException('Syntax error: "' + i_value + '" is not a valid value for ' + ParameterKey.Name + '.');
  end
  else
  begin
    // TODO Control, whether NumberStyles.Float is appropriate
    if Double.TryParse(i_value, NumberStyles.Float,CultureInfo.InstalledUICulture, out i_fake) then
      Add(new ConstantParameter(ParameterKey, Convert.ToDouble(i_value, CultureInfo.InstalledUICulture)))
    else
      Add(new ConstantParameter(ParameterKey, Convert.ToDouble(i_value, CultureInfo.InvariantCulture)));
  end;
end;

constructor Parameters(NumberOfYears, NumberOfRegions: Integer; rand: Random);
begin
  inherited constructor;
  m_numberOfRegions:=NumberOfRegions;
  m_numberOfYears:=NumberOfYears;
  UseReflectionToCreateObjects;
  m_globalParameters:=new Hashtable;
  m_regionParameters:=new Hashtable;
  m_regionRegionParameters:=new Dictionary<String, Parameter>;
  m_yearParameters:=new Hashtable;
  m_yearRegionParameters:=new Hashtable;
  m_allParameterNameAndId:=new List<ParameterIdAndName>;
  _rand:=rand;
end;

function Parameters.GetItem(Name: string): double;
begin
  if m_globalParameters.ContainsKey(Name) then
    Result:=Parameter(m_globalParameters.Item[Name]).Value
  else
    raise new Exception('Parameter with the name "' + Name + '" does not exist in the input file');
end;

function Parameters.GetItemByRegion(Region, Name: string): double;
begin
  if m_RegionParameters.ContainsKey(Region + ';' + Name) then
    Result:=Parameter(m_RegionParameters.Item[Region + ';' + Name]).Value
  else
    raise new Exception('Parameter with the name "' + Name + '" for Region "' + Region + '" does not exist in the input file');
end;

function Parameters.GetItemByYear(Year: integer; Name: string): double;
begin
  if m_yearParameters.ContainsKey(Convert.ToString(Year) + ';' + Name) then
    Result:=Parameter(m_yearParameters.Item[Convert.ToString(Year) + ';' + Name]).Value
  else
    // TODO -oDavid Anthoff: Change 1950 to constant
    raise new ApplicationException('Parameter with the name "' + Name + '" for Year ' + Convert.ToString(Year+1950) + ' does not exist in the input file');
end;

function Parameters.GetItemByYearAndRegion(Year: integer; Region,
  Name: string): double;
begin
  if m_yearRegionParameters.ContainsKey(Convert.ToString(Year) + ';' + Region + ';' + Name) then
    Result:=Parameter(m_yearRegionParameters.Item[Convert.ToString(Year) + ';' + Region + ';' + Name]).Value
  else
    raise new ApplicationException('Parameter with the name "' + Name + '" for Region "' + Region + '" for the year ' + Convert.ToString(Year + 1950) + ' does not exist in the input file');
end;

procedure Parameters.ReadExcelFile(FileName: string);
begin
  if Filename.EndsWith('.xls') then
  begin
    ReadExcel2003File(FileName);
  end
  else if Filename.EndsWith('.xlsm') then
  begin
    ReadExcel2007File(FileName);
  end
  else
  begin
    raise new Exception('Unknown file format');
  end;
end;

procedure Parameters.SetToBestGuessValue;
var
  i_enumerator: IEnumerator;
begin
  i_enumerator:=m_globalParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToBestGuessValue;
  end;

  i_enumerator:=m_regionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToBestGuessValue;
  end;
  
  i_enumerator:=m_regionRegionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToBestGuessValue;
  end;

  i_enumerator:=m_yearParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToBestGuessValue;
  end;

  i_enumerator:=m_yearRegionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToBestGuessValue;
  end;

  UseReflectionToSetFieldsCurrentValues;
end;

procedure Parameters.SetToRandomValue;
var
  i_enumerator: IEnumerator;
begin
  i_enumerator:=m_globalParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToRandomValue;
  end;

  i_enumerator:=m_regionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToRandomValue;
  end;
  
  i_enumerator:=m_regionRegionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToRandomValue;
  end;

  i_enumerator:=m_yearParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToRandomValue;
  end;

  i_enumerator:=m_yearRegionParameters.Values.GetEnumerator;
  while i_enumerator.MoveNext do
  begin
    Parameter(i_enumerator.Current).SetToRandomValue;
  end;

  UseReflectionToSetFieldsCurrentValues;
end;

procedure Parameters.UseReflectionToCreateObjects;
var
  i_Type: System.Type;
  i_Fields: array of System.Reflection.FieldInfo;
  i: integer;
  i_attributes: array of Object;
begin
  i_Type:=Self.GetType;
  i_Fields:=i_Type.GetFields;

  for i:=0 to High(i_Fields) do
  begin
    i_attributes:=i_fields[i].GetCustomAttributes(typeof(ParameterAttribute), false);
    if Length(i_attributes)=1 then
    begin
      if i_fields[i].FieldType.Name='RegionParameter' then
      begin
        i_fields[i].SetValue(Self, new RegionParameter);
      end
      else if i_fields[i].FieldType.Name='RegionRegionParameter' then
      begin
        i_fields[i].SetValue(Self, new RegionRegionParameter);
      end
      else if i_fields[i].FieldType.Name='YearParameter' then
      begin
        i_fields[i].SetValue(Self, new YearParameter);
      end
      else if i_fields[i].FieldType.Name='YearRegionParameter' then
      begin
        i_fields[i].SetValue(Self, new YearRegionParameter);
      end;
    end
    else if Length(i_attributes)>1 then
      raise new Exception('More then one ParameterAttribute specified for "' + i_Fields[i].Name + '".');
  end;
end;

procedure Parameters.UseReflectionToSetFieldsCurrentValues;
var
  i_Type: System.Type;
  i_Fields: array of System.Reflection.FieldInfo;
  i, l, k: integer;
  i_attributes: array of Object;
  i_RegionParameter: RegionParameter;
  i_YearParameter: YearParameter;
  i_YearRegionParameter: YearRegionParameter;
begin
  i_Type:=Self.GetType;
  i_Fields:=i_Type.GetFields;

  for i:=0 to High(i_Fields) do
  begin
    i_attributes:=i_fields[i].GetCustomAttributes(typeof(ParameterAttribute), false);
    if Length(i_attributes)=1 then
    begin
      if i_fields[i].FieldType.Name='Double' then
      begin
        if ParameterAttribute(i_attributes[0]).Name=nil then
          i_fields[i].SetValue(Self, Object(GetItem(i_fields[i].Name)))
        else
          i_fields[i].SetValue(Self, Object(GetItem(ParameterAttribute(i_attributes[0]).Name)));
      end
      else if i_fields[i].FieldType.Name='RegionParameter' then
      begin
        i_RegionParameter:=RegionParameter(i_fields[i].GetValue(Self));
        if ParameterAttribute(i_attributes[0]).Name=nil then
        begin
          for l:=1 to LegacyConstants.NoReg do
            i_RegionParameter.Item[l]:=GetItemByRegion(LegacyRegionNames.Name[l], i_fields[i].Name);
        end
        else
        begin
          for l:=1 to LegacyConstants.NoReg do
            i_RegionParameter.Item[l]:=GetItemByRegion(LegacyRegionNames.Name[l], ParameterAttribute(i_attributes[0]).Name);
        end;
      end
      else if i_fields[i].FieldType.Name='RegionRegionParameter' then
      begin
        var i_RegionRegionParameter:=RegionRegionParameter(i_fields[i].GetValue(Self));
        if ParameterAttribute(i_attributes[0]).Name=nil then
        begin
          for l:=1 to LegacyConstants.NoReg do
            for k:=1 to LegacyConstants.NoReg do
              i_RegionRegionParameter[l, k]:=GetItemByRegionRegion(LegacyRegionNames.Name[l], LegacyRegionNames.Name[k], i_fields[i].Name);
        end
        else
        begin
          for l:=1 to LegacyConstants.NoReg do
            for k:=1 to LegacyConstants.NoReg do
              i_RegionRegionParameter[l, k]:=GetItemByRegionRegion(LegacyRegionNames.Name[l], LegacyRegionNames.Name[k], ParameterAttribute(i_attributes[0]).Name);
        end;
      end
      else if i_fields[i].FieldType.Name='YearParameter' then
      begin
        i_YearParameter:=YearParameter(i_fields[i].GetValue(Self));
        if ParameterAttribute(i_attributes[0]).Name=nil then
        begin
          for l:=0 to LegacyConstants.NYear do
            i_YearParameter.Item[l]:=GetItemByYear(l, i_fields[i].Name);
        end
        else
        begin
          for l:=0 to LegacyConstants.NYear do
            i_YearParameter.Item[l]:=GetItemByYear(l, ParameterAttribute(i_attributes[0]).Name);
        end;
      end
      else if i_fields[i].FieldType.Name='YearRegionParameter' then
      begin
        i_YearRegionParameter:=YearRegionParameter(i_fields[i].GetValue(Self));
        if ParameterAttribute(i_attributes[0]).Name=nil then
        begin
          for l:=0 to LegacyConstants.NYear do
            for k:=1 to LegacyConstants.NoReg do
              i_YearRegionParameter.Item[l, k]:=GetItemByYearAndRegion(l, LegacyRegionNames.Name[k], i_fields[i].Name);
        end
        else
        begin
          for l:=0 to LegacyConstants.NYear do
            for k:=1 to LegacyConstants.NoReg do
              i_YearRegionParameter.Item[l, k]:=GetItemByYearAndRegion(l, LegacyRegionNames.Name[k], ParameterAttribute(i_attributes[0]).Name);
        end;
      end;
    end
    else if Length(i_attributes)>1 then
      raise new Exception('More then one ParameterAttribute specified for "' + i_Fields[i].Name + '".');
  end;
end;

procedure Parameters.WriteCurrentValuesToFile(FileName: string);
var
  i_writer: TextWriter;
  i_enumerator: IEnumerator;
begin
  i_writer:=new StreamWriter(FileName);
  try
    i_enumerator:=m_globalParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      i_writer.Write(Parameter(i_enumerator.Current).Key.Name);
      i_writer.Write(' = ');
      i_writer.WriteLine(Parameter(i_enumerator.Current).Value);
    end;

    i_enumerator:=m_regionParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      i_writer.Write(Parameter(i_enumerator.Current).Key.Name);
      i_writer.Write(' (' + RegionParameterKey(Parameter(i_enumerator.Current).Key).RegionName + ')');
      i_writer.Write(' = ');
      i_writer.WriteLine(Parameter(i_enumerator.Current).Value);
    end;
    
    i_enumerator:=m_regionRegionParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      i_writer.Write(Parameter(i_enumerator.Current).Key.Name);
      i_writer.Write(' (' + RegionRegionParameterKey(Parameter(i_enumerator.Current).Key).Region1 + ';' + RegionRegionParameterKey(Parameter(i_enumerator.Current).Key).Region2 + ')');
      i_writer.Write(' = ');
      i_writer.WriteLine(Parameter(i_enumerator.Current).Value);
    end;

    i_enumerator:=m_yearParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      i_writer.Write(Parameter(i_enumerator.Current).Key.Name);
      i_writer.Write(' (' + Convert.ToString(YearParameterKey(Parameter(i_enumerator.Current).Key).Year) + ')');
      i_writer.Write(' = ');
      i_writer.WriteLine(Parameter(i_enumerator.Current).Value);
    end;

    i_enumerator:=m_yearRegionParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      i_writer.Write(Parameter(i_enumerator.Current).Key.Name);
      i_writer.Write(' (');
      i_writer.Write(Convert.ToString(YearRegionParameterKey(Parameter(i_enumerator.Current).Key).Year));
      i_writer.Write(';');
      i_writer.Write(YearRegionParameterKey(Parameter(i_enumerator.Current).Key).Region + ')');
      i_writer.Write(' = ');
      i_writer.WriteLine(Parameter(i_enumerator.Current).Value);
    end;
  finally
    i_writer.Close;
  end;
end;

procedure Parameters.WriteValuesToCsv (ScenarioId: Integer; RunId: Integer;
                                GlobalCsv, RegionCsv , YearCsv, RegionYearCsv: TextWriter);
var
  i_enumerator : IEnumerator;
  lparameter : Parameter;
begin
  if GlobalCsv <> nil then
  begin
    i_enumerator:=m_globalParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      lparameter:=Parameter(i_enumerator.Current);
      GlobalCsv.Write(ScenarioId);
      GlobalCsv.Write(';');
      GlobalCsv.Write(RunId);
      GlobalCsv.Write(';');
      GlobalCsv.Write(lparameter.Key.Id);
      GlobalCsv.Write(';');
      GlobalCsv.Write(lparameter.Value);
      GlobalCsv.WriteLine;
    end;
  end;

  if RegionCsv <> nil then
  begin
    i_enumerator:=m_regionParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      lparameter:=Parameter(i_enumerator.Current);
      RegionCsv.Write(ScenarioId);
      RegionCsv.Write(';');
      RegionCsv.Write(RunId);
      RegionCsv.Write(';');
      RegionCsv.Write(lparameter.Key.Id);
      RegionCsv.Write(';');
      RegionCsv.Write(RegionParameterKey(lparameter.Key).RegionId);
      RegionCsv.Write(';');
      RegionCsv.Write(lparameter.Value);
      RegionCsv.WriteLine;
    end;
  end;
  
  // TODO: RegionRegionCsv

  if YearCsv <> nil then
  begin
    i_enumerator:=m_yearParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      lparameter:=Parameter(i_enumerator.Current);
      YearCsv.Write(ScenarioId);
      YearCsv.Write(';');
      YearCsv.Write(RunId);
      YearCsv.Write(';');
      YearCsv.Write(lparameter.Key.Id);
      YearCsv.Write(';');
      YearCsv.Write(YearParameterKey(lparameter.Key).Year + 1950);
      YearCsv.Write(';');
      YearCsv.Write(lparameter.Value);
      YearCsv.WriteLine;
    end;
  end;

  if RegionYearCsv <> nil then
  begin
    i_enumerator:=m_yearRegionParameters.Values.GetEnumerator;
    while i_enumerator.MoveNext do
    begin
      lparameter:=Parameter(i_enumerator.Current);
      RegionYearCsv.Write(ScenarioId);
      RegionYearCsv.Write(';');
      RegionYearCsv.Write(RunId);
      RegionYearCsv.Write(';');
      RegionYearCsv.Write(lparameter.Key.Id);
      RegionYearCsv.Write(';');
      RegionYearCsv.Write(YearRegionParameterKey(lparameter.Key).RegionId);
      RegionYearCsv.Write(';');
      RegionYearCsv.Write(YearRegionParameterKey(lparameter.Key).Year + 1950);
      RegionYearCsv.Write(';');
      RegionYearCsv.Write(lparameter.Value);
      RegionYearCsv.WriteLine;
    end;
  end;
end;

method Parameters.GetItemByRegionRegion(region1, region2, name: string): double; 
begin
  var lParameterKey:=String.Format('{0};{1};{2}', region1, region2, name);
  
  if m_regionRegionParameters.ContainsKey(lParameterKey) then
    Result:=m_regionRegionParameters[lParameterKey].Value
  else
    raise new ApplicationException(String.Format('Parameter with name "{0}" for region1 "{1}" and region2 "{2}" does not exist in the input files', name, region1, region2));
end;

method Parameters.ReadExcel2003File(Filename: string); 
var
  i, l: integer;
  lConnection: OleDbConnection;
  lCmd: OleDbCommand;
  lReader: OleDbDataReader;
  lRegionNamesForSQL: string;
  lYearlyParameterNamesForSQL: string;
  lTables: DataTable;
  lFilterColumns: array [0..2] of Object;
  i_TableName: string;
  lParameterId: Integer := 0;
begin
  try
    Console.Write('Loading parameters from Excel file');

    Console.Write('.');
    lConnection:=new OleDbConnection;
    lConnection.ConnectionString:='Provider=Microsoft.Jet.OLEDB.4.0;Data Source=' + FileName + ';Extended Properties="Excel 8.0;IMEX=1";';
    lConnection.Open;
    try
      lCmd:=lConnection.CreateCommand;
      lCmd.CommandText:='SELECT Name, Value FROM [Global$]';
      lReader:=lCmd.ExecuteReader;
      try
        while lReader.Read do
        begin
          var lNameAndId := new ParameterIdAndName;
          lNameAndId.Name:=Convert.ToString(lReader.Item['Name']).Trim();

          if m_allParameterIdsByName.ContainsKey(lNameAndId.Name) then
            lNameAndId.Id:=m_allParameterIdsByName[lNameAndId.Name]
          else
          begin
            lNameAndId.Id:=lParameterId;
            m_allParameterIdsByName.Add(lNameAndId.Name, lNameAndId.Id);
            Inc(lParameterId);
          end;
          
          m_allParameterNameAndId.Add(lNameAndId);
          Add(new GlobalParameterKey(lNameAndId), Convert.ToString(lReader.Item['Value']));
        end;
      finally
        lReader.Close;
      end;

      lRegionNamesForSQL:='';
      for i:=1 to m_numberOfRegions do
      begin
        lRegionNamesForSQL:=lRegionNamesForSQL + ', ' + LegacyRegionNames.Name[i];
      end;

      Console.Write('.');
      lCmd:=lConnection.CreateCommand;
      lCmd.CommandText:='SELECT Name' + lRegionNamesForSQL + ' FROM [Regional$]';
      lReader:=lCmd.ExecuteReader;
      try
        while lReader.Read do
        begin
          if not lReader.IsDBNull(lReader.GetOrdinal('Name')) then
            if lReader.Item['Name'].ToString.Trim().Length>0 then
            begin
              var lNameAndId := new ParameterIdAndName;
              lNameAndId.Name:=Convert.ToString(lReader.Item['Name']).ToString();
              
              if m_allParameterIdsByName.ContainsKey(lNameAndId.Name) then
                lNameAndId.Id:=m_allParameterIdsByName[lNameAndId.Name]
              else
              begin
                lNameAndId.Id:=lParameterId;
                m_allParameterIdsByName.Add(lNameAndId.Name, lNameAndId.Id);
                Inc(lParameterId);
              end;
              
              m_allParameterNameAndId.Add(lNameAndId);
              for i:=1 to LegacyConstants.NoReg do
              begin
                Add(new RegionParameterKey(LegacyRegionNames.Name[i], lNameAndId), Convert.ToString(lReader.Item[LegacyRegionNames.Name[i]]));
              end;
            end;
        end;
      finally
        lReader.Close;
      end;

      var lyearlyParameterNamesAndIds := new List<ParameterIdAndName>;
      lFilterColumns[0]:=nil;
      lFilterColumns[1]:=nil;
      lFilterColumns[2]:='Yearly$';
      lTables:=lConnection.GetOleDbSchemaTable(OleDbSchemaGuid.Columns, lFilterColumns);
      lYearlyParameterNamesForSQL:='';
      for i:=0 to lTables.Rows.Count - 1 do
      begin
        if lTables.Rows.Item[i].Item[3].ToString<>'Year' then
        begin
          var lNameAndId := new ParameterIdAndName;
          lNameAndId.Name:=lTables.Rows.Item[i].Item[3].ToString;
          
          if m_allParameterIdsByName.ContainsKey(lNameAndId.Name) then
            lNameAndId.Id:=m_allParameterIdsByName[lNameAndId.Name]
          else
          begin
            lNameAndId.Id:=lParameterId;
            m_allParameterIdsByName.Add(lNameAndId.Name, lNameAndId.Id);
            Inc(lParameterId);
          end;
          
          m_allParameterNameAndId.Add(lNameAndId);
          lyearlyParameterNamesAndIds.Add(lNameAndId);
        
          lYearlyParameterNamesForSQL:=lYearlyParameterNamesForSQL + ', [' + lTables.Rows.Item[i].Item[3].ToString + ']';
        end;
      end;

      Console.Write('.');
      lCmd:=lConnection.CreateCommand;
      lCmd.CommandText:='SELECT Year' + lYearlyParameterNamesForSQL + ' FROM [Yearly$]';
      lReader:=lCmd.ExecuteReader;
      try
        while lReader.Read do
        begin
          for i:=0 to lTables.Rows.Count - 1 do
          begin
            if lTables.Rows.Item[i].Item[3].ToString<>'Year' then
            begin
          
              // TODO -oDavid Anthoff: Change 1950 to a constant
              Add(new YearParameterKey(Convert.ToInt32(lReader.Item['Year'])-1950, lyearlyParameterNamesAndIds[i]), Convert.ToString(lReader.Item[lTables.Rows.Item[i].Item[3].ToString]));
            end;
          end;
        end;
      finally
        lReader.Close;
      end;

      Console.Write('.');
      lTables:=lConnection.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, nil);
      for i:=0 to lTables.Rows.Count - 1 do
      begin
        i_TableName:=lTables.Rows.Item[i].Item[2].ToString;
        if i_TableName.EndsWith('$') then
        begin
          i_TableName:=i_TableName.Substring(0, i_TableName.IndexOf('$'));
          if (i_TableName<>'Global') and (i_TableName<>'Regional') and (i_TableName<>'Yearly') then
          begin
            var lNameAndId := new ParameterIdAndName;
            lNameAndId.Name:=i_TableName;
              
            if m_allParameterIdsByName.ContainsKey(lNameAndId.Name) then
              lNameAndId.Id:=m_allParameterIdsByName[lNameAndId.Name]
            else
            begin
              lNameAndId.Id:=lParameterId;
              m_allParameterIdsByName.Add(lNameAndId.Name, lNameAndId.Id);
              Inc(lParameterId);
            end;
            
            m_allParameterNameAndId.Add(lNameAndId);
          
            if i_TableName.ToLower.StartsWith('rr') then
            begin
              lCmd:=lConnection.CreateCommand;
              lCmd.CommandText:='SELECT Region' + lRegionNamesForSQL + ' FROM [' + i_TableName + '$]';
              lReader:=lCmd.ExecuteReader;
              try
                for ii: integer:=1 to LegacyConstants.NoReg do
                begin
                  lReader.Read;
                  for l:=1 to LegacyConstants.NoReg do
                  begin
                    Add(new RegionRegionParameterKey(LegacyRegionNames.Name[l], Convert.ToString(lReader.Item['Region']), lNameAndId), Convert.ToString(lReader.Item[LegacyRegionNames.Name[l]]));
                  end;
                end;
              finally
                lReader.Close;
              end;
            end
            else
            begin
              lCmd:=lConnection.CreateCommand;
              lCmd.CommandText:='SELECT Year' + lRegionNamesForSQL + ' FROM [' + i_TableName + '$]';
              lReader:=lCmd.ExecuteReader;
              try
                while lReader.Read do
                begin
                  for l:=1 to LegacyConstants.NoReg do
                  begin
                    Add(new YearRegionParameterKey(Convert.ToInt32(lReader.Item['Year'])-1950, LegacyRegionNames.Name[l], lNameAndId), Convert.ToString(lReader.Item[LegacyRegionNames.Name[l]]));
                  end;
                end;
              finally
                lReader.Close;
              end;
            end;
          end;
        end;
      end;
    finally
      lConnection.Close;
    end;
  finally
    Console.WriteLine;
  end;
end;

method Parameters.ReadExcel2007File(Filename: string); 
var
  i, l: integer;
  lConnection: OleDbConnection;
  lCmd: OleDbCommand;
  lReader: OleDbDataReader;
  lRegionNamesForSQL: string;
  lYearlyParameterNamesForSQL: string;
  lTables: DataTable;
  lFilterColumns: array [0..2] of Object;
  i_TableName: string;
  lParameterId: Integer := 0;
begin
  try
    Console.Write('Loading parameters from Excel file');
    
    var excelParameters:=new Esmf.FundExcelParameterFile(Filename);
    excelParameters.Load;
    
    for each p: Esmf.FundExcelParameterFile.Parameter in excelParameters.Parameters do
    begin
      var lNameAndId := new ParameterIdAndName;
      lNameAndId.Name:=p.Name;

      if m_allParameterIdsByName.ContainsKey(lNameAndId.Name) then
        lNameAndId.Id:=m_allParameterIdsByName[lNameAndId.Name]
      else
      begin
        lNameAndId.Id:=lParameterId;
        m_allParameterIdsByName.Add(lNameAndId.Name, lNameAndId.Id);
        Inc(lParameterId);
      end;
      
      m_allParameterNameAndId.Add(lNameAndId);
        
      if p is Esmf.FundExcelParameterFile.NonDimensionalParameter then
      begin
        var pp:=Esmf.FundExcelParameterFile.NonDimensionalParameter(p);
        Add(new GlobalParameterKey(lNameAndId), pp.Value);
      end
      else if p is Esmf.FundExcelParameterFile.OneDimensionalParameter then
      begin
        var pp:=Esmf.FundExcelParameterFile.OneDimensionalParameter(p);
        
        if pp.Length=16 then
        begin
          for i:=1 to LegacyConstants.NoReg do
          begin
            Add(new RegionParameterKey(LegacyRegionNames.Name[i], lNameAndId), pp[i-1]);
          end;          
        end
        else
        begin
          for i:=0 to LegacyConstants.NYear do
          begin
            Add(new YearParameterKey(i, lNameAndId), pp[i]);
          end;          
        end;        
      end
      else if p is Esmf.FundExcelParameterFile.TwoDimensionalParameter then
      begin
        var pp:=Esmf.FundExcelParameterFile.TwoDimensionalParameter(p);
        
        if pp.RowCount>16 then
        begin
          for i:=1 to LegacyConstants.NoReg do
          begin
            for l:=0 to LegacyConstants.NYear do
            begin
              Add(new YearRegionParameterKey(l, LegacyRegionNames.Name[i], lNameAndId), pp[i-1,l]);
            end;                     
          end;          
          
        end
        else
        begin
          for i:=1 to LegacyConstants.NoReg do
          begin
            for l:=1 to LegacyConstants.NoReg do
            begin
              Add(new RegionRegionParameterKey(LegacyRegionNames.Name[i], LegacyRegionNames.Name[l], lNameAndId), pp[i-1,l-1]);
            end;                     
          end;          
        end;
      end;
    end;
  finally
    Console.WriteLine;
  end;
end;

{ ConstantParameter }

constructor ConstantParameter(Key: ParameterKey; Value: double);
begin
  inherited constructor(Key, nil);
  m_value:=Value;
end;

function ConstantParameter.GetValue: double;
begin
  Result:=m_value;
end;

procedure ConstantParameter.SetToBestGuessValue;
begin

end;

procedure ConstantParameter.SetToRandomValue;
begin

end;

{ NormalDistributionParameter }

constructor NormalDistributionParameter(Key: ParameterKey; Arguments: array of string; rand: Random);
begin
  inherited constructor(Key, rand);
  m_bestGuess:=Convert.ToDouble(Arguments[0], CultureInfo.InvariantCulture);
  m_standardDeviation:=Convert.ToDouble(Arguments[1], CultureInfo.InvariantCulture);

  if High(Arguments)>1 then
  begin
    if Arguments[2].Trim().Length>0 then
    begin
      m_useLowerBound:=true;
      m_lowerBound:=Convert.ToDouble(Arguments[2], CultureInfo.InvariantCulture);
    end
    else
      m_useLowerBound:=false;

    if High(Arguments)>2 then
    begin
      if Arguments[3].Trim().Length>0 then
      begin
        m_useUpperBound:=true;
        m_upperBound:=Convert.ToDouble(Arguments[3], CultureInfo.InvariantCulture);
      end
      else
        m_useUpperBound:=false;
    end;
  end;

  SetToBestGuessValue;
end;

constructor NormalDistributionParameter(Key: ParameterKey;
  BestGuess, StandardDeviation: double; rand: Random);
begin
  inherited constructor(Key, rand);
  m_bestGuess:=BestGuess;
  m_standardDeviation:=StandardDeviation;
  m_useLowerBound:=false;
  m_useUpperBound:=false;
  SetToBestGuessValue;
end;

constructor NormalDistributionParameter(Key: ParameterKey;
  BestGuess, StandardDeviation: double; UseLowerBound: Boolean;
  LowerBound: double; UseUpperBound: Boolean; UpperBound: double; rand: Random);
begin
  inherited constructor(Key, rand);
  m_bestGuess:=BestGuess;
  m_standardDeviation:=StandardDeviation;
  m_useLowerBound:=UseLowerBound;
  m_useUpperBound:=UseUpperBound;
  m_lowerBound:=LowerBound;
  m_upperBound:=UpperBound;
  SetToBestGuessValue;
end;

function NormalDistributionParameter.GetValue: double;
begin
  Result:=m_value;
end;

procedure NormalDistributionParameter.SetToBestGuessValue;
begin
  m_value:=m_bestGuess;
end;

procedure NormalDistributionParameter.SetToRandomValue;
var
  i_helper1, i_helper2: double;
begin
  LegacyFund28Math.GenNormal(m_bestGuess, m_standardDeviation, var i_helper1, var i_helper2, _rand);
  m_value:=i_helper1;

  while (m_useLowerBound and (m_value<m_lowerBound)) or (m_useUpperBound and (m_value>m_UpperBound)) do
  begin
    LegacyFund28Math.GenNormal(m_bestGuess, m_standardDeviation, var i_helper1, var i_helper2, _rand);
    m_value:=i_helper1;
  end;
end;

{ TriangularDistributionParameter }

constructor TriangularDistributionParameter(Key: ParameterKey; Arguments: array of string; rand: Random);
begin
  inherited constructor(Key, rand);
  m_min:=Convert.ToDouble(Arguments[0], CultureInfo.InvariantCulture);
  m_bestGuess:=Convert.ToDouble(Arguments[1], CultureInfo.InvariantCulture);
  m_max:=Convert.ToDouble(Arguments[2], CultureInfo.InvariantCulture);
  SetToBestGuessValue;
end;

function TriangularDistributionParameter.GetValue: double;
begin
  Result:=m_value;
end;

procedure TriangularDistributionParameter.SetToBestGuessValue;
begin
  m_value:=m_bestGuess;
end;

procedure TriangularDistributionParameter.SetToRandomValue;
var
  i_helper: double;
begin
  LegacyFund28Math.GenTriang(m_min, m_max, m_bestGuess, var i_helper, _rand);
  m_value:=i_helper;
end;

{ GlobalParameterKey }

constructor GlobalParameterKey(NameAndId: ParameterIdAndName);
begin
  inherited constructor(NameAndId);
end;

{ RegionParameterKey }

constructor RegionParameterKey(Region: string; NameAndId: ParameterIdAndName);
begin
  inherited constructor(NameAndId);
  m_region:=Region;
  m_regionId:=-1;
  
  for i: Integer := 1 to LegacyConstants.NoReg do
  begin
    if LegacyRegionNames.Name[i]=Region then
    begin
      m_regionId:=i;
      break;
    end; 
  end;
  
  if m_regionId=-1 then
    raise new ApplicationException('Region not found.');
end;

{ YearParameterKey }

constructor YearParameterKey(YearData: integer; NameAndId: ParameterIdAndName);
begin
  inherited constructor(NameAndId);
  m_year:=YearData;
end;

{ YearRegionParameterKey }

constructor YearRegionParameterKey(YearData: integer; RegionData: string; 
  NameAndId: ParameterIdAndName);
begin
  inherited constructor(NameAndId);
  m_year:=YearData;
  m_region:=RegionData;
  
  m_regionId:=-1;
  for i: Integer := 1 to LegacyConstants.NoReg do
  begin
    if LegacyRegionNames.Name[i]=RegionData then
    begin
      m_regionId:=i;
      break;
    end; 
  end;
  
  if m_regionId=-1 then
    raise new ApplicationException('Region not found.');
  
end;

{ ParameterKey }

constructor ParameterKey(NameAndId: ParameterIdAndName);
begin
  inherited constructor;
  m_NameAndId:=NameAndId;
end;

method ParameterKey.GetName: String;
begin
  Result:=m_NameAndId.Name;
end;

method ParameterKey.GetId: Integer;
begin
  Result:=m_NameAndId.Id;
end;


{ Parameter }

constructor Parameter(KeyData: ParameterKey; rand: Random);
begin
  inherited constructor;
  m_key:=KeyData;
  _rand:=rand;
end;

{ ExponentialDistributionParameter }

constructor ExponentialDistributionParameter(Key: ParameterKey;
  Arguments: array of string; rand: Random);
begin
  inherited constructor(Key, rand);
  m_lambda:=Convert.ToDouble(Arguments[0], CultureInfo.InvariantCulture);
  m_useLowerBound:=false;
  m_useUpperBound:=false;

  if High(Arguments)>0 then
  begin
    if Arguments[1].Trim().Length>0 then
    begin
      m_useLowerBound:=true;
      m_lowerBound:=Convert.ToDouble(Arguments[1], CultureInfo.InvariantCulture);
    end;

    if High(Arguments)>1 then
    begin
      if Arguments[2].Trim().Length>0 then
      begin
        m_useUpperBound:=true;
        m_upperBound:=Convert.ToDouble(Arguments[2], CultureInfo.InvariantCulture);
      end;
    end;
  end;

  SetToBestGuessValue;
end;

function ExponentialDistributionParameter.GetValue: double;
begin
  Result:=m_value;
end;

procedure ExponentialDistributionParameter.SetToBestGuessValue;
begin
  m_value:=0;
end;

procedure ExponentialDistributionParameter.SetToRandomValue;
var
  i_helper: double;
begin
  LegacyFund28Math.GenExponential(m_lambda, var i_helper, _rand);
  m_value:=i_helper;

  while (m_useLowerBound and (m_value<m_lowerBound)) or (m_useUpperBound and (m_value>m_UpperBound)) do
  begin
    LegacyFund28Math.GenExponential(m_lambda, var i_helper, _rand);
    m_value:=i_helper;
  end;

end;

{ GammaDistributionParameter }

constructor GammaDistributionParameter(Key: ParameterKey;
  Arguments: array of string; rand: Random);
begin
  inherited constructor(Key, rand);
  m_alpha:=Convert.ToDouble(Arguments[0], CultureInfo.InvariantCulture);
  m_beta:=Convert.ToDouble(Arguments[1], CultureInfo.InvariantCulture);
  SetToBestGuessValue;
end;

function GammaDistributionParameter.GetValue: double;
begin
  Result:=m_value;
end;

procedure GammaDistributionParameter.SetToBestGuessValue;
begin
  m_value:=m_beta*(m_alpha-1);
end;

procedure GammaDistributionParameter.SetToRandomValue;
var
  i_helper: double;
begin
  LegacyFund28Math.GenGamma(m_alpha, m_beta, var i_helper, _rand);
  m_value:=i_helper;
end;

{ YearParameter }

constructor YearParameter;
begin
  inherited constructor;
end;

function YearParameter.GetItem(Year: integer): double;
begin
  Result:=m_values[Year];
end;

procedure YearParameter.SetItem(Year: integer; Value: double);
begin
  m_values[Year]:=Value;
end;

{ RegionParameter }

constructor RegionParameter;
begin
  inherited constructor;
end;

function RegionParameter.GetItem(Region: integer): double;
begin
  Result:=m_Values[Region-1];
end;

procedure RegionParameter.SetItem(Region: integer; Value: double);
begin
  m_values[Region-1]:=Value;
end;

{ YearRegionParameter }

constructor YearRegionParameter;
begin
  inherited constructor;
end;

function YearRegionParameter.GetItem(Year, Region: integer): double;
begin
  Result:=m_values[Year, Region - 1];
end;

procedure YearRegionParameter.SetItem(Year, Region: integer;
  Value: double);
begin
  m_values[Year, Region - 1]:=Value;
end;

constructor ParameterAttribute; 
begin
end;

procedure RegionRegionParameter.SetItem(region1: Integer; region2: integer; value: double); 
begin
  m_values[region1-1, region2-1]:=value;
end;

function RegionRegionParameter.GetItem(region1: integer; region2: integer): double; 
begin
  Result:=m_Values[region1-1, region2-1];
end;

constructor RegionRegionParameter; 
begin
  inherited;
end;

constructor RegionRegionParameterKey(region1Data, region2Data: string; NameAndId: ParameterIdAndName); 
begin
  inherited constructor(NameAndId);
  m_region1:=region1Data;
  m_region1Id:=-1;
  m_region2:=region2Data;
  m_region2Id:=-1;
  
  for i: Integer := 1 to LegacyConstants.NoReg do
  begin
    if LegacyRegionNames.Name[i]=region1Data then
    begin
      m_region1Id:=i;
    end; 
    
    if LegacyRegionNames.Name[i]=region2Data then
    begin
      m_region2Id:=i;
    end;
  end;
  
  if (m_region1Id=-1) or (m_region2Id=-1) then
    raise new ApplicationException('Region not found.');
end;

end.
