// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

uses
  CommandLine;

type
  
  FundArguments = class
  public
    [DefaultArgument(ArgumentType.AtMostOnce)]
    ConfigurationFilename: string;  
    [Argument(ArgumentType.AtMostOnce)]
    Quiet: boolean;    
  end;
  
  ConsoleApp = class
  private
    class method ShowVersionInfo;
    class method UnhandledExceptionHandler(sender: object; args: UnhandledExceptionEventArgs);  
  public
    class TempOutputFile: System.IO.TextWriter;
    class method Main(args: array of string);
  end;
  
implementation

uses
    System.Threading,
    System.Configuration,
    System.Globalization,
    System.IO,
    System.Collections,
    System.Collections.Generic,
    System.Diagnostics;

class method ConsoleApp.ShowVersionInfo;
begin
  Console.WriteLine('FUND - Climate Framework for Uncertainty, Negotiation and Distribution');
  Console.WriteLine('Version ' + System.Reflection.Assembly.GetExecutingAssembly.GetName.Version.ToString);
  Console.WriteLine;
end;


class method ConsoleApp.Main(args: array of string);
var
  lMarginalRun: TMarginalRun;

  gDisaggregatedCsvFile: TextWriter;
  gDisaggregatedYearCsvFile: Textwriter;
  gDisaggregatedYearRegionCsvFile: Textwriter;
  gDisaggregatedYearRegionWeightingschemeCsvFile: Textwriter;
  gBGEResultsCSVFile: TextWriter;

  lAggregateMarginalDamage: TextWriter;

  lDimGasCsv: TextWriter;
  lDimRunCsv: TextWriter;
  lDimRegionCsv: TextWriter;
  lDimSectorCsv: TextWriter;
  lDimYearCsv: TextWriter;
  lGlobalInputCsv: TextWriter;
  lRegionInputCsv: TextWriter;
  lYearInputCsv: TextWriter;
  lRegionYearInputCsv: TextWriter;
  lParameterIdCsv: TextWriter;
  lYearStr: string;
  lRuns: Integer;
  lOutputPath: String;
  lParam: FundParameter;
  lRandom: Random;
  
  lSimulationManager: SimulationManager;
  lParsedCmdArgs: FundArguments;
begin
    var appd:=AppDomain.CurrentDomain;
    appd.UnhandledException+=new UnhandledExceptionEventHandler(ConsoleApp.UnhandledExceptionHandler);
        
    // Set the thread to the invariant culture, so that Write and WriteLn
    // format all floating point variables with a . as the decimal separator
    // That way the Windows setting for locals is overriden

    Thread.CurrentThread.CurrentCulture:=CultureInfo.InvariantCulture;

    // Set thread priority to below normal, so that other programs can be used
    // while FUND is running
    Thread.CurrentThread.Priority:=ThreadPriority.BelowNormal;

    ShowVersionInfo;

    lParsedCmdArgs:=new FundArguments;
    if Parser.ParseArgumentsWithUsage(args, lParsedCmdArgs) then
    begin          
      
      var lDefaultConfigurationFile:=Path.Combine(Path.Combine(Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly.Location), 'Data'), 'DefaultSimulation.xml');
      lSimulationManager:=new SimulationManager(iif(lParsedCmdArgs.ConfigurationFilename=nil, lDefaultConfigurationFile,lParsedCmdArgs.ConfigurationFilename));
      lSimulationManager.Load;
      
      lOutputPath:=Path.Combine(Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly.Location), 'Data');
      
      if not Directory.Exists(lOutputPath) then
        Directory.CreateDirectory(lOutputPath);
      
      case lSimulationManager.Rng of
        RandomNumberGenerator.DotNet: 
        begin
          if lSimulationManager.Randomize then
            lRandom:=new Random()
          else
            lRandom:=new Random(34514325);    
        end;
        RandomNumberGenerator.MersenneTwister: 
        begin
          if lSimulationManager.Randomize then
          begin
            raise new Exception('Support for randomization is not there for mersenne twister');
          end  
          else
            lRandom:=new jp.takel.PseudoRandom.MersenneTwister();
        end;
        else raise new Exception('You specified a RNG that is not implemented');
      end;

      lAggregateMarginalDamage:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Aggregate marginal damage.csv'));
      gDisaggregatedYearCsvFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact Year.csv'));
      TempOutputFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Temp.csv'));
      gBGEResultsCSVFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - BGE.csv'));
      
      if lSimulationManager.OutputVerbal then
      begin
        lAggregateMarginalDamage.Write('Scenario');
        lAggregateMarginalDamage.Write(';');
        lAggregateMarginalDamage.Write('Gas');
        lAggregateMarginalDamage.Write(';');
        lAggregateMarginalDamage.Write('Emissionyear');
        lAggregateMarginalDamage.Write(';');
        lAggregateMarginalDamage.Write('Run');
        lAggregateMarginalDamage.Write(';');
        lAggregateMarginalDamage.Write('Weightingscheme');
        lAggregateMarginalDamage.Write(';');
        lAggregateMarginalDamage.Write('Marginal damage');
        lAggregateMarginalDamage.WriteLine;
      end;

      if lSimulationManager.OutputDisaggregatedData then
      begin
        gDisaggregatedCsvFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact YearRegionSectorWeightingscheme.csv'));
        gDisaggregatedYearRegionCsvFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact YearRegionOutput.csv'));
        gDisaggregatedYearRegionWeightingschemeCsvFile:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact YearRegionWeightingschemeOutput.csv'));
        
        if lSimulationManager.OutputVerbal then
        begin
          gDisaggregatedCsvFile.Write('Scenario');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Run');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Marginal Gas');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Marginal Emission Year');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Year');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Region');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Sector');
          gDisaggregatedCsvFile.Write(';');          
          gDisaggregatedCsvFile.Write('Weightingscheme');
          gDisaggregatedCsvFile.Write(';');
          gDisaggregatedCsvFile.Write('Damage');
          gDisaggregatedCsvFile.WriteLine;
        end;
      end;
      
      lDimGasCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Gas.csv'));
      lDimGasCsv.WriteLine('0;C');
      lDimGasCsv.WriteLine('1;CH4');
      lDimGasCsv.WriteLine('2;N2O');
      lDimGasCsv.WriteLine('3;SF6');
      lDimGasCsv.Close;

      lDimRunCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Run.csv'));
      lDimRunCsv.WriteLine('0;Best guess');
      for i: integer:=1 to lRuns do
      begin
        lDimRunCsv.Write(i);
        lDimRunCsv.Write(';');
        lDimRunCsv.Write(i);
        lDimRunCsv.WriteLine;
      end;
      lDimRunCsv.Close;

      lDimRegionCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Region.csv'));
      for i: integer:=1 to LegacyConstants.NoReg do
      begin
        lDimRegionCsv.WriteLine(i.ToString + ';' + LegacyRegionNames.Name[i] + '');
      end;
      lDimRegionCsv.Close;

      lDimSectorCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Sector.csv'));
      lDimSectorCsv.WriteLine('0;eloss;Water');
      lDimSectorCsv.WriteLine('1;eloss;Forests');
      lDimSectorCsv.WriteLine('2;eloss;Heating');
      lDimSectorCsv.WriteLine('3;eloss;Cooling');
      lDimSectorCsv.WriteLine('4;eloss;Agriculture');
      lDimSectorCsv.WriteLine('5;eloss;Dryland');
      lDimSectorCsv.WriteLine('6;eloss;SeaProtection');
      lDimSectorCsv.WriteLine('7;eloss;Imigration');
      lDimSectorCsv.WriteLine('8;sloss;Species');
      lDimSectorCsv.WriteLine('9;sloss;Death');
      lDimSectorCsv.WriteLine('10;sloss;Morbidity');
      lDimSectorCsv.WriteLine('11;sloss;Wetland');
      lDimSectorCsv.WriteLine('12;sloss;Emigration');
      lDimSectorCsv.WriteLine('13;eloss;Hurricane');

      lDimSectorCsv.Close;


      using lDimScenarioCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Scenario.csv')) do
      begin
        for each lScenario: Scenario in lSimulationManager.Scenarios do
        begin
          // Write Scenario dimension file
          lDimScenarioCsv.Write(lScenario.Id);
          lDimScenarioCsv.Write(';');
          lDimScenarioCsv.Write(lScenario.Name);
          lDimScenarioCsv.WriteLine;
        end;
      end;
      
      
      lDimYearCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Year.csv'));
      for i: Integer:=1950 to 2300 do
      begin
        lYearStr := i.ToString;
        lDimYearCsv.Write(lYearStr);
        lDimYearCsv.Write(';');
        lDimYearCsv.Write(lYearStr.Substring(0,2));
        lDimYearCsv.Write('xx;');
        lDimYearCsv.Write(lYearStr.Substring(0,3));
        lDimYearCsv.Write('x;');
        lDimYearCsv.Write(lYearStr.Substring(0,4));
        lDimYearCsv.WriteLine();
      end;
      lDimYearCsv.Close;
      
      var lDimEmissionYear:=new List<integer>;
      
      if lSimulationManager.OutputInputParameters then
      begin
        lGlobalInputCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact Parameter.csv'));
        lRegionInputCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact ParameterRegion.csv'));
        lYearInputCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact ParameterYear.csv'));
        lRegionYearInputCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Fact ParameterYearRegion.csv'));
        lParameterIdCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Parameter.csv'));
      end;

      for each lRun: Run in lSimulationManager.Runs do
      begin
        lParam:=new FundParameter(LegacyConstants.NYear, LegacyConstants.NoReg, lRandom);
        for each filename: string in lRun.Scenario.ExcelFiles do
        begin
          lParam.ReadExcelFile(filename);         
        end;

        Console.WriteLine(lRun.Scenario.Name);
        // Set defaults
        lRuns:=lRun.MonteCarloRuns;
        
        if lRun.Scengmt=true then
          raise new NotImplementedException('scengmt support is not working right now');

        if lRun.Mode=RunMode.MarginalRun then
        begin     
          // DA: Use MargMain for marginal cost, use Main for total cost and optimisation
          // modes
          lMarginalRun:=new TMarginalRun(lRun, MarginalGas.C, 2005, lOutputPath, lParam);
          
          if not lDimEmissionYear.Contains(2005) then
            lDimEmissionYear.Add(2005);
          
          lMarginalRun.AggregateDamageCsv:=lAggregateMarginalDamage;

          if lSimulationManager.OutputDisaggregatedData then
          begin
            lMarginalRun.YearRegionCsv:=gDisaggregatedYearRegionCsvFile;
            lMarginalRun.YearRegionWeightingSchemeCsv:=gDisaggregatedYearRegionWeightingschemeCsvFile;
            lMarginalRun.YearRegionSectorWeightingSchemeCsv:=gDisaggregatedCsvFile;
          end;
          
          if lSimulationManager.OutputInputParameters then
          begin
            lMarginalRun.GlobalInputCsv:=lGlobalInputCsv;
            lMarginalRun.RegionInputCsv:=lRegionInputCsv;
            lMarginalRun.YearInputCsv:=lYearInputCsv;
            lMarginalRun.RegionYearInputCsv:=lRegionYearInputCsv;
          end;    
          
          var watch:=new System.Diagnostics.Stopwatch();
          watch.Start();
          lMarginalRun.Run;
          watch.Stop();
          //Console.WriteLine('Elapsed time: {0}', watch.Elapsed);
        end
        else if lRun.Mode=RunMode.OptimalTaxation then
        begin
          var prtps:array of Double:=[0.03,0.01,0.001];
          var etas:Array of Double:=[1.0];
          var negishis: Array of Boolean:=[true,false];
          
          using fileOptTax:=System.IO.File.CreateText(Path.Combine(lOutputPath,'Output - Optimal Tax.csv')) do
          begin
            using fileTemp:=System.IO.File.CreateText(Path.Combine(lOutputPath,'Output - Temperature.csv')) do
            begin
              using fileRadforc:=System.IO.File.CreateText(Path.Combine(lOutputPath,'Output - Radforce.csv')) do
              begin
                for each prtp: Double in prtps do
                begin
                  for each eta: Double in etas do
                  begin
                    for each negishi: Boolean in negishis do
                    begin
                      Console.WriteLine('Now doing eta {0} prtp {1} negishi {2}', eta, prtp, negishi);
                      var lOptimalTaxation:=new OptimalTaxation(
                        Prtp:=prtp,
                        Eta:=eta,
                        UseNegishiWeights := negishi,
                        Parameters := lParam);
                      
                        lOptimalTaxation.PostRunMethod:=method(model: FundWorkflow);
                          begin
                            Esmf.OutputHelper.WriteParameterToFile(model.SocioEconomicState.currtax, fileOptTax); 
                            Esmf.OutputHelper.WriteParameterToFile(model.ClimateState.temp, fileTemp);
                            Esmf.OutputHelper.WriteParameterToFile(model.ClimateState.radforc, fileRadforc);
                          end;
                      lOptimalTaxation.Run;                        
                      
                    end;
                  end;
                end;                                      
              end;
            end;
          end;


        end
        else if lRun.Mode=RunMode.MarginalRun2 then
        begin
          var prtps:array of Double:=[0.03,0.01,0.001];
          var etas:Array of Double:=[1.0,1.5,2.0];
          
          for each prtp: Double in prtps do
          begin
            for each eta: Double in etas do
            begin
              var lMarginalRun2:=new MarginalDamage2(
                Run:=lRun,
                Gas:=MarginalGas.C,
                EmissionYear:=Esmf.Timestep.FromYear(2005),
                UseNegishiWeights:=true,
                Parameters:=lParam,
                Eta:=eta,
                Prtp:=prtp);
              var scc:=lMarginalRun2.Start;
              Console.WriteLine('prtp {0}, eta {1}, scc {2:f2}', prtp, eta, scc[Fund.CommonDimensions.Region.USA]);
            end;
          end;
          
//          for each r: Fund.CommonDimensions.Region in Enum.GetValues(typeof(Fund.CommonDimensions.Region)) do
//          begin
//            Console.WriteLine('r: {0}, SCC: ${1:f2}', r, scc[r]);
//          end;
          
        end
        else if lRun.Mode=RunMode.BGERun then
        begin
          // DA: Use MargMain for marginal cost, use Main for total cost and optimisation
          // modes
          var lBGERun:=new TBGERun(lRun, lParam);
          lBGERun.BGEResultsCsv:=gBGEResultsCSVFile;
                 
          var watch:=new System.Diagnostics.Stopwatch();
          watch.Start();
          lBGERun.Run;
          watch.Stop();
          //Console.WriteLine('Elapsed time: {0}', watch.Elapsed);
        end
        else if lRun.Mode=RunMode.FullMarginalRun then
        begin     
          for each gas: MarginalGas in Enum.GetValues(typeof(MarginalGas)) do
          begin
            for emissionyear: Integer:=2005 to 2095 step 10 do
            begin 
              if not lDimEmissionYear.Contains(emissionyear) then
                 lDimEmissionYear.Add(emissionyear);
              
              Console.WriteLine("Now doing year {0} and gas {1}", emissionyear, gas);             
              // DA: Use MargMain for marginal cost, use Main for total cost and optimisation
              // modes
              lMarginalRun:=new TMarginalRun(lRun, gas, emissionyear, lOutputPath, lParam);
              
              lMarginalRun.AggregateDamageCsv:=lAggregateMarginalDamage;

              if lSimulationManager.OutputDisaggregatedData then
              begin
                lMarginalRun.YearRegionCsv:=gDisaggregatedYearRegionCsvFile;
                lMarginalRun.YearRegionWeightingSchemeCsv:=gDisaggregatedYearRegionWeightingschemeCsvFile;
                lMarginalRun.YearRegionSectorWeightingSchemeCsv:=gDisaggregatedCsvFile;
              end;
              
              if lSimulationManager.OutputInputParameters then
              begin
                lMarginalRun.GlobalInputCsv:=lGlobalInputCsv;
                lMarginalRun.RegionInputCsv:=lRegionInputCsv;
                lMarginalRun.YearInputCsv:=lYearInputCsv;
                lMarginalRun.RegionYearInputCsv:=lRegionYearInputCsv;
              end;    
              
              var watch:=new System.Diagnostics.Stopwatch();
              watch.Start();
              lMarginalRun.Run;
              watch.Stop();
              //Console.WriteLine('Elapsed time: {0}', watch.Elapsed);
            end;
          end;
        end
        else if lRun.Mode=RunMode.EpaMarginalRun then
        begin     
          var emissionyears:array of Integer:=[2005,2030];
                    
          for each gas: MarginalGas in Enum.GetValues(typeof(MarginalGas)) do
          begin            
            for each emissionyear: Integer in emissionyears do
            begin 
              if not lDimEmissionYear.Contains(emissionyear) then
                 lDimEmissionYear.Add(emissionyear);
              
              Console.WriteLine("Now doing year {0} and gas {1}", emissionyear, gas);             
              // DA: Use MargMain for marginal cost, use Main for total cost and optimisation
              // modes
              lMarginalRun:=new TMarginalRun(lRun, gas, emissionyear, lOutputPath, lParam);
              
              lMarginalRun.AggregateDamageCsv:=lAggregateMarginalDamage;

              if lSimulationManager.OutputDisaggregatedData then
              begin
                lMarginalRun.YearRegionCsv:=gDisaggregatedYearRegionCsvFile;
                lMarginalRun.YearRegionWeightingSchemeCsv:=gDisaggregatedYearRegionWeightingschemeCsvFile;
                lMarginalRun.YearRegionSectorWeightingSchemeCsv:=gDisaggregatedCsvFile;
              end;
              
              if lSimulationManager.OutputInputParameters then
              begin
                lMarginalRun.GlobalInputCsv:=lGlobalInputCsv;
                lMarginalRun.RegionInputCsv:=lRegionInputCsv;
                lMarginalRun.YearInputCsv:=lYearInputCsv;
                lMarginalRun.RegionYearInputCsv:=lRegionYearInputCsv;
              end;    
              
              var watch:=new System.Diagnostics.Stopwatch();
              watch.Start();
              lMarginalRun.Run;
              watch.Stop();
              //Console.WriteLine('Elapsed time: {0}', watch.Elapsed);
            end;
          end;
        end
        else if lRun.Mode=RunMode.TotalRun then
        begin
          // DA: Use MargMain for marginal cost, use Main for total cost and optimisation
          // modes
          var lTotalDamageRun:=new TotalDamage(lRun, lOutputPath, lParam);
          
          lTotalDamageRun.AggregateDamageCsv:=lAggregateMarginalDamage;

          if lSimulationManager.OutputDisaggregatedData then
          begin
            lTotalDamageRun.YearRegionCsv:=gDisaggregatedYearRegionCsvFile;
            lTotalDamageRun.YearRegionWeightingSchemeCsv:=gDisaggregatedYearRegionWeightingschemeCsvFile;
            lTotalDamageRun.YearRegionSectorWeightingSchemeCsv:=gDisaggregatedCsvFile;
          end;
          
          if lSimulationManager.OutputInputParameters then
          begin
            lTotalDamageRun.GlobalInputCsv:=lGlobalInputCsv;
            lTotalDamageRun.RegionInputCsv:=lRegionInputCsv;
            lTotalDamageRun.YearInputCsv:=lYearInputCsv;
            lTotalDamageRun.RegionYearInputCsv:=lRegionYearInputCsv;
          end;      

          lTotalDamageRun.Run;
        
        end
        else if lRun.Mode=RunMode.CEARun then
        begin
          var lCEARun:=new TCEARun(lRun, lParam);
          
          lCEARun.CEA;
        end;
      end;

      lAggregateMarginalDamage.Close;
      gDisaggregatedYearCsvFile.Close;
      TempOutputFile.Close;
      gBGEResultsCSVFile.Close;
      
      using lDimEmissionYearCsv:=new StreamWriter(Path.Combine(lOutputPath, 'Output - Dim Emissionyear.csv')) do
      begin
        for each emissionyear: integer in lDimEmissionYear do
        begin
          lDimEmissionYearCsv.WriteLine('{0};{1}', emissionyear, emissionyear);       
        end;
      end;

      if lSimulationManager.OutputDisaggregatedData then
      begin
        gDisaggregatedCsvFile.Close;
        gDisaggregatedYearRegionCsvFile.Close;
        gDisaggregatedYearRegionWeightingschemeCsvFile.Close;
      end;
      
      if lSimulationManager.OutputInputParameters then
      begin
        lGlobalInputCsv.Close;
        lRegionInputCsv.Close;
        lYearInputCsv.Close;
        lRegionYearInputCsv.Close;

        for each param: ParameterIdAndName in lParam.ParameterIdAndNames do
        begin
          lParameterIdCsv.Write(param.Id);
          lParameterIdCsv.Write(';');
          lParameterIdCsv.Write(param.Name);
          lParameterIdCsv.WriteLine;      
        end;

        lParameterIdCsv.Close;
      end;    

      Console.WriteLine;
      Console.WriteLine;
      Console.WriteLine('Finished');
               
      if not lParsedCmdArgs.Quiet then
      begin
        Console.WriteLine('Press Enter to quit');
        Console.ReadLine;
      end;
    end;
end;

class method ConsoleApp.UnhandledExceptionHandler(sender: object; args: UnhandledExceptionEventArgs); 
begin
  var e:=Exception(args.ExceptionObject);
  
  Console.WriteLine('An unhandled exception occurred that is handled by the the ConsoleApp.UnhandledExceptionHandler:');
  Console.WriteLine;
  Console.WriteLine(e.GetType().Name);
  Console.WriteLine(e.Message);
  Console.WriteLine(e.Source);
  Console.WriteLine(e.StackTrace);
end;

end.