// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;
//  Last Changed: 21 January 2006

interface

type
  
  LegacyConstants = public static class
  private
  public
    const NYear = 349;      // number of years
    const NoReg = 16;       // number of regions
    const NoSector = 14;    // number of sectors
    const BaseT = 3.0;      // base case warming
    const DBsT  = 0.04;     // base case yearly warming
    const BaseS = 0.50;     // base case sea level rise T=2.5
    const DBsS  = 0.0083;   // base case yearly rise    
  end;
    
  // Enum that can be used to categorise damage. If the number of options here
  // is changed, make sure to also adjust NoSector!
  Sector = public (sWater,
                   sForests, 
                   sHeating, 
                   sCooling, 
                   sAgriculture,
                   sDryland, 
                   sSeaProtection, 
                   sImigration, 
                   sSpecies, 
                   sDeath, 
                   sMorbidity, 
                   sWetland, 
                   sEmigration,
                   sHurrican);
                   
  Year = array [0..LegacyConstants.NYear] of double;
 
  RegArr = array [1..LegacyConstants.NoReg] of double;
  
  RegRec = public class
  public
    income: double;           // income in billion US 1995 $ ?!? Might also be %
    oldincome: double;        // income in the previous period
    ecgrsd: double;
    population: double;       // population in millions
    YpC: double;              // income per capita
    urbpop: double;           // urban population
    urbpop0: double;
    aeeisd: double;
    aceisd: double;
    reei: double;             // reduction energy efficiency
    seei: double;
    emission: double;         // emissions from fossil fuel
    emiss0: double;           // emissions in base year
    cummemm: double;
    foremsd: double;
    forestry: Year;         // reduction through forestry
    forcost: double;          // forestry costs

    dead: double;             // number of deaths
    deadcost: double;
    morbcost: double;
    yll: double;              // years of life lost
    yld: double;              // years of life diseased
    species: double;          // loss species
    water: double;            // water
    forests: double;          // forestry
    heating: double;          // heating
    cooling: double;          // cooling
    hurrdam: double;            // hurricane damage
    //hurrbasedam: double;        // hurricane damage, 1990
    //hurrdamel: double;         // income elasticity
    hurrdead: double;          // hurricane deaths
    //hurrbasedead: double;      // hurricane deaths, 1990
    //hurrdeadel: double;        // income elasticity
    //hurrpar: Double;           // increase in hurricane intensity per degree warming
    //hurrnl: Double;            // non-linearity of damage increase
    dengue: double;           // dengue fever
    schisto: double;          // schistosomiasis
    malaria: double;          // malaria
    diadead: double;          // diarrhea
    diasick: double;
    agcost: double;           // agriculture, general
    plus: double;
    basecardvasc: double;
    cardcold: double;         // cardiovascular, cold

    cardheat: double;         // cardiovascular, heat
    resp: double;             // respiratory
    baseresp: double;
    wetland: double;          // wetlands
    wetmeas: double;
    wettot: double;
    wetann: double;
    wetval: double;
    npwetcost: double;
    wetcost: double;
    dryland: double;          // dryland
    dryval: double;
    npdrycost: double;
    drycost: double;
    protlev: double;          // protection level
    protcost: double;         // protection costs
    npprotcost: double;
    popdens: double;          // population density
    area: double;
    leave: double;
    cumleave: double;
    enter: double;
    cumenter: double;
    leavecost: double;
    entercost: double;

    // parameters
    sf6: double;                   
    regtmp: double;
    regdt: double;
    
    // The index of that region, which is just the position in
    // the array EconArr.
    regionIndex: integer;
    cumaeei: double;
  end;
 
  ClimateClass = public class
  public
    acco2: Year;      // carbon dioxide
    sea: double;      // sea level
    ds: double;       // sea difference
  end;

  FundParameter = public class( Parameters )
  protected
    method Getvaladj(region: Integer): double;
    method Getch4decay: double;
    method Getn2odecay: double;
    method Getsf6decay: Double;
    method Getdecay(index: Integer): double;
    method Getdelaysea: double;
    method Getdelaytemp: double;
    method Gettemps: double;
    method Getseas: double;
    method Getimmigrate(region1, region2: integer): double;  
    method Getypc90(region: integer): double;  
  public
    // time you must host refugees before economically back on their feet
    [ParameterAttribute]
    entertime: double;
    
    // n2o uncertainty
    [ParameterAttribute]
    n2oadd: double;

    // ch4 uncertainty
    [ParameterAttribute]
    ch4add: double;

    // life ch4
    [ParameterAttribute]
    lifech4: double;

    // life n20
    [ParameterAttribute]
    lifen2o: double;

    // life sea
    [ParameterAttribute]
    lifesea: double;

    // share of agriculture
    [ParameterAttribute]
    agrish0: RegionParameter;

    // base regional temperature parameters
    [ParameterAttribute]
    bregtmp: RegionParameter;
    
    [ParameterAttribute]
    temp90: RegionParameter;

    // Extra methane emissions for Megan's extreme scenarios
    [ParameterAttribute]
    RunAwayCH4: YearParameter;

    // income growth per year
    // economic growth
    [ParameterAttribute]
    ypcgrowth: YearRegionParameter;

    // population growth per year
    [ParameterAttribute]
    pgrowth: YearRegionParameter;
    
    [ParameterAttribute]
    gdp90: RegionParameter;
    
    [ParameterAttribute]
    pop90: RegionParameter;

    // emission from deforestation, MtCO2
    [ParameterAttribute(Name:='ForestEmm')]
    forestemm: YearRegionParameter;

    // autonomous energy efficiency improvement
    [ParameterAttribute(Name:='AEEI')]
    aeei: YearRegionParameter;

    // autonomous carbon efficiency improvements
    [ParameterAttribute(Name:='ACEI')]
    acei: YearRegionParameter;

    // exogenous methane emissions
    [ParameterAttribute]
    ch4em: YearRegionParameter;

    // Davids aggregation equity weights
    [ParameterAttribute(Name:='ewAggregationFactor')]
    davideqf: YearRegionParameter;

    // benchmark water resources costs
    [ParameterAttribute]
    wrbm: RegionParameter;

    // benchmark heating costs
    [ParameterAttribute]
    hebm: RegionParameter;

    // benchmark cooling costs
    [ParameterAttribute]
    cebm: RegionParameter;
   
    // hurricane damage
    [ParameterAttribute]
    hurrbasedam: RegionParameter;

    // income elasticity
    [ParameterAttribute]
    hurrdamel: double;

    // hurricane casualties
    [ParameterAttribute]
    hurrbasedead: RegionParameter;

    // income elasticity
    [ParameterAttribute]
    hurrdeadel: double;

    // climate sensitivity
    [ParameterAttribute]
    hurrpar: double;

    // climate non-linearity
    [ParameterAttribute]
    hurrnl: double;
        
    // benchmark forestry costs
    [ParameterAttribute]
    forbm: RegionParameter;

    // benchmark dengue fever
    [ParameterAttribute]
    dfbs: RegionParameter;
    
    [ParameterAttribute]
    dfch: RegionParameter;

    // benchmark schistosomiasis
    [ParameterAttribute]
    smbs: RegionParameter;
    
    [ParameterAttribute]
    smch: RegionParameter;

    // benchmark malaria
    [ParameterAttribute]
    malbs: RegionParameter;
    
    [ParameterAttribute]
    malch: RegionParameter;

    // life-losses from cardheat.
    [ParameterAttribute]
    chmlbm: RegionParameter;
    
    [ParameterAttribute]
    chmqbm: RegionParameter;
    
    [ParameterAttribute]
    chplbm: RegionParameter;
    
    [ParameterAttribute]
    chpqbm: RegionParameter;

    //life-losses from cardcold
    [ParameterAttribute]
    ccmlbm: RegionParameter;
    
    [ParameterAttribute]
    ccmqbm: RegionParameter;
    
    [ParameterAttribute]
    ccplbm: RegionParameter;
    
    [ParameterAttribute]
    ccpqbm: RegionParameter;

    // life-losses from resp
    [ParameterAttribute]
    rlbm: RegionParameter;
    
    [ParameterAttribute]
    rqbm: RegionParameter;

    // wetland losses sea level rise
    [ParameterAttribute]
    wlbm: RegionParameter;
    
    [ParameterAttribute]
    wmbm: RegionParameter;
    
    [ParameterAttribute]
    wetmax: RegionParameter;

    // dryland losses sea level rise
    [ParameterAttribute]
    dlbm: RegionParameter;

    // protection costs sea level rise 
    [ParameterAttribute]
    coast: RegionParameter;
    
    [ParameterAttribute]
    pc: RegionParameter;

    // agricultural losses
    [ParameterAttribute]
    agrbm: RegionParameter;
    
    [ParameterAttribute]
    aglbm: RegionParameter;
    
    [ParameterAttribute]
    agopt: RegionParameter;
    
    [ParameterAttribute]
    agtime: RegionParameter;
    
    [ParameterAttribute(Name:='agcbm')]
    agcbmInternal: RegionParameter;

    // uncertainty about scenarios
    [ParameterAttribute]
    pgadd: RegionParameter;
    
    [ParameterAttribute]
    ecgradd: RegionParameter;
    
    [ParameterAttribute]
    aeeiadd: RegionParameter;
    
    [ParameterAttribute]
    aceiadd: RegionParameter;
    
    [ParameterAttribute]
    foremadd: RegionParameter;

    // benchmark dengue fever
    // Not uncertain, because they are data
    [ParameterAttribute]
    diamort: RegionParameter;
    
    [ParameterAttribute]
    diayld: RegionParameter;

    // reduction cost parameter for forestry
    // forestry cost constant
    [ParameterAttribute]
    forconst: RegionParameter;
    
    // forestry cost linear
    [ParameterAttribute]
    forlinear: RegionParameter;
    
    // forestry cost square
    [ParameterAttribute]
    forsquare: RegionParameter;

    (* reads base income from inc0.001 *)
    // initial income
    [ParameterAttribute(Name:='GDP0')]
    gdp0: RegionParameter;

    // initial SO2
    [ParameterAttribute]
    so20: RegionParameter;

    // methane parameters
    [ParameterAttribute]
    ch4par1: RegionParameter;
    
    [ParameterAttribute]
    ch4par2: RegionParameter;

    // exogenous n2o emissions
    [ParameterAttribute]
    n2oem: YearRegionParameter;

    // nitrous oxide parameters
    [ParameterAttribute]
    n2opar1: RegionParameter;
    
    [ParameterAttribute]
    n2opar2: RegionParameter;
    
    // SF6 emissions, 1990
    [ParameterAttribute]
    sf60: RegionParameter;
    
    // SF6 parameters
    [ParameterAttribute]
    sf6ypc: Double;
    
    [ParameterAttribute]
    sf6gdp: Double;
    
    [ParameterAttribute]
    lifesf6: Double;
    
    [ParameterAttribute]
    sf6add: Double;
    
    // exogenous temperatures 
    [ParameterAttribute]
    scentemp: YearRegionParameter;

    // dead-to-years-of-life-disabled transformation parameters
    [ParameterAttribute]
    d2dm: RegionParameter;
    
    [ParameterAttribute]
    d2ds: RegionParameter;
    
    [ParameterAttribute]
    d2dd: RegionParameter;
    
    [ParameterAttribute]
    d2dc: RegionParameter;
    
    [ParameterAttribute]
    d2dr: RegionParameter;

    // dead-to-years-of-life-lost transformation parameters
    [ParameterAttribute]
    d2lm: RegionParameter;
    
    [ParameterAttribute]
    d2ls: RegionParameter;
    
    [ParameterAttribute]
    d2ld: RegionParameter;
    
    [ParameterAttribute]
    d2lc: RegionParameter;
    
    [ParameterAttribute]
    d2lr: RegionParameter;

    // Discount rate. Not every where used!!!
    [ParameterAttribute]
    discrate: RegionParameter;

    // initial population
    [ParameterAttribute]
    pop0: RegionParameter;

    [ParameterAttribute]
    bioloss: double;
    
    [ParameterAttribute]
    biosens: double;
    
    [ParameterAttribute]    
    bioshare: double;
    
    [ParameterAttribute]
    modet: double;
    
    [ParameterAttribute]
    modes: double;  // modal sea level rise T=1
    
    [ParameterAttribute]
    watechrate: double;
    
    [ParameterAttribute]
    forel: double;
    
    [ParameterAttribute]
    fornl: double;
    
    [ParameterAttribute]
    wrel: double;

    [ParameterAttribute]
    wrnl: double;

    [ParameterAttribute]
    heel: double;

    [ParameterAttribute]
    henl: double;

    [ParameterAttribute]
    ceel: double;

    [ParameterAttribute]
    cenl: double;

    [ParameterAttribute]
    agnl: double;

    [ParameterAttribute]
    agel: double;

    [ParameterAttribute]
    malnl: double;

    [ParameterAttribute]
    dfnl: double;

    [ParameterAttribute]
    smnl: double;

    [ParameterAttribute]
    forco2: double;

    [ParameterAttribute]
    plusel: double;

    [ParameterAttribute]
    spbm: double;

    [ParameterAttribute]
    vsl: double;

    [ParameterAttribute]
    vmorb: double;

    [ParameterAttribute]
    vyll: double;

    [ParameterAttribute]
    diamortel: double;

    [ParameterAttribute]
    diamortnl: double;

    [ParameterAttribute]
    diayldel: double;

    [ParameterAttribute]
    diayldnl: double;

    [ParameterAttribute]
    emcst: double;

    [ParameterAttribute]
    immcst: double;

    [ParameterAttribute]
    cvlin: double;

    [ParameterAttribute]
    rlin: double;

    [ParameterAttribute]
    maxcardvasc: double;

    [ParameterAttribute]
    leavedel: double;

    [ParameterAttribute]
    vbel: double;

    [ParameterAttribute]
    lifeco1: double;

    [ParameterAttribute]
    lifeco2: double;

    [ParameterAttribute]
    lifeco3: double;

    [ParameterAttribute]
    lifeco4: double;

    [ParameterAttribute]
    lifeco5: double;

    [ParameterAttribute]
    lifetemp: double;

    [ParameterAttribute]
    valbase: double;

    [ParameterAttribute]
    valinc: RegionParameter;

    [ParameterAttribute]
    wvbm: RegionParameter;

    [ParameterAttribute]
    dvbm: RegionParameter;

    [ParameterAttribute]
    wetdel: RegionParameter;

    [ParameterAttribute]
    drydel: RegionParameter;

    [ParameterAttribute]
    taxmp: RegionParameter;

    [ParameterAttribute]
    formp: RegionParameter;
    
    [ParameterAttribute]
    DBsTa: double;
    
    [ParameterAttribute]
    nospecbase: double;

    // Jiehan's prtp
    [ParameterAttribute]
    prtp: RegionParameter;

    // Which gas is used for the marginal
    // evaluation (0=CO2, 1=CH4, 2=N20)
    [ParameterAttribute(Name:='marginalgas')]
    marginalGas: double;

    // In which year is the emission
    // of the marginal gases done?
    [ParameterAttribute(Name:='emissionperiod')]
    emissionPeriod: double;
    
    [ParameterAttribute(Name:='Migrate')]
    migrate: RegionRegionParameter;
    
    [ParameterAttribute]
    co2pre: double;
    
    [ParameterAttribute]
    emissint0: RegionParameter;
    
    [Parameter]
    urbcorr: RegionParameter;
    
    // pop dens at coast, relative to total
    [Parameter]
    coastpd: RegionParameter;
    
    [Parameter]
    plus90: RegionParameter;  
    
    [Parameter]
    cardvasc90: RegionParameter;

    [Parameter]
    resp90: RegionParameter;

    [Parameter]
    incdens: Double;
    
    property valadj[Region: Integer]: double read Getvaladj;
    property ch4decay: double read Getch4decay;
    property n2odecay: double read Getn2odecay;
    property sf6decay: double read Getsf6decay;
    property delaysea: double read Getdelaysea;
    property delaytemp: double read Getdelaytemp;
    property decay[Index: Integer]: double read Getdecay;
    
    // Temperature sensitivity
    property temps: double read Gettemps;
    
    // sea level sensitivity
    property seas : double read Getseas;
    
    property immigrate[region1, region2: integer]: double read Getimmigrate;   
    property ypc90[region: Integer]: double read Getypc90; 
  end;

implementation

method FundParameter.Getch4decay: double;
begin
  Result:=1/lifech4;
end;

method FundParameter.Getn2odecay: double;
begin
  Result:=1/lifen2o;
end;

method FundParameter.Getsf6decay: double;
begin
  Result:=1/lifesf6;
end;

method FundParameter.Getdelaysea: double;
begin
  Result:=1/lifesea;
end;

method FundParameter.Gettemps: double;
begin
  Result:=modet/5.35/LegacyFund28Math.ln(2);
end;

method FundParameter.Getseas: double;
begin
  Result:=modes;
end;

method FundParameter.Getdecay(Index: Integer): double;
begin
  case Index of
    1: Result:=lifeco1;
    2: Result:=LegacyFund28Math.Exp(-1/lifeco2);
    3: Result:=LegacyFund28Math.Exp(-1/lifeco3);
    4: Result:=LegacyFund28Math.Exp(-1/lifeco4);
    5: Result:=LegacyFund28Math.Exp(-1/lifeco5);
    else
      raise new ArgumentOutOfRangeException;
  end;
end;

method FundParameter.Getdelaytemp: double;
begin
  Result:=1/lifetemp;
end;

method FundParameter.Getvaladj(Region: Integer): double;
begin
  Result:=valbase/valinc[Region]/(1+valbase/valinc[Region]);
end;

method FundParameter.Getimmigrate(region1, region2: integer): double;
var
  immsumm: double:=0;
begin
  for i: integer:=1 to LegacyConstants.NoReg do
    immsumm:=immsumm + migrate[region1, i];
    
  Result:=migrate[region1, region2]/immsumm;
end;

method FundParameter.Getypc90(region: integer): double;  
begin
  Result:=gdp90[region]/pop90[region]*1000;
end;

end.
