// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

uses
  System.Configuration,
  System.IO,
  Esmf,
  System.Collections.Generic;

type
  
  ModelModifierDelegate = method(fundModel: FundWorkflow); 
  
  BGESimulationCombination = class
  public
    property prtp: Double;
    property eta: Double;
    property taxscenarios: array of double;
  end;
  
  BGE = class
  public
    property GlobalBGE: Double;
    property RegionalBGE: Double; 
    
    property ExpectedTempIn2050: Double;
    property ExpectedTempIn2100: Double; 
    property MaxExpectedTemp: Double;

    property ExpectedRfIn2050: Double;
    property ExpectedRfIn2100: Double; 
    property MaxExpectedRf: Double;     
    
    property TrimmedForNaNGlobalWelfare: integer;
    property TrimmedForNaNRegionalWelfare: Integer;
    property TrimmedForNaNTempIn2050: Integer;
    property TrimmedForNaNTempIn2100: Integer;
    property TrimmedForNaNRfIn2050: Integer;
    property TrimmedForNaNRfIn2100: Integer;
    property TrimmedForNaNTempMax: Integer;
    property TrimmedForNaNRfMax: Integer;   
  end;   
  
  TBGERun = class
  private
    { Private Declarations }
    
    _run: Run;
    _parameters: FundParameter;
        
    method CalculateBGE(preRunModifications: ModelModifierDelegate; eta, prtp: Double): BGE;
    method OptimumFinder(maxNumberOfSteps: Integer; eta, prtp: double): IEnumerable<KeyValuePair<double,BGE>>; iterator;
  public
    constructor(runData: Run; parameter: FundParameter);

    method Run;
    
    property BGEResultsCsv: Textwriter;    
  end;

implementation

constructor TBGERun(runData: Run; parameter: FundParameter);
begin
  _parameters:=parameter;
  _run:=runData;
end;

method TBGERun.Run;
begin
  var scenarios:=new List<BGESimulationCombination>;
  
  scenarios.Add(new BGESimulationCombination(prtp:=0.03,eta:=1.0,taxscenarios:=[0.0{,1.0,2.0,3.0,4.0,5.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.03,eta:=1.5,taxscenarios:=[0.0{,1.0,2.0,3.0,4.0,5.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.03,eta:=2.0,taxscenarios:=[0.0{,1.0,2.0,3.0,4.0,5.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.01,eta:=1.0,taxscenarios:=[0.0{,5.0,10.0,15.0,20.0,25.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.01,eta:=1.5,taxscenarios:=[0.0{,5.0,10.0,15.0,20.0,25.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.01,eta:=2.0,taxscenarios:=[0.0{,5.0,10.0,15.0,20.0,25.0}]));
  scenarios.Add(new BGESimulationCombination(prtp:=0.001,eta:=1.0,taxscenarios:=[0.0{,20.0,40.0,50.0,60.0,80.0,100.0}]));      
  scenarios.Add(new BGESimulationCombination(prtp:=0.001,eta:=1.5,taxscenarios:=[0.0{,20.0,40.0,50.0,60.0,80.0,100.0}]));      
  scenarios.Add(new BGESimulationCombination(prtp:=0.001,eta:=2.0,taxscenarios:=[0.0{,20.0,40.0,50.0,60.0,80.0,100.0}]));      
     
  var bgeresults:=new List<string>;
  
  for each scenario in scenarios do
  begin     
    var bgebytaxes:=new Dictionary<double,BGE>;
      
    Console.WriteLine;
    Console.WriteLine("With damage");
    Console.WriteLine;
    
    var bgewithdamage:=CalculateBGE(
      method(fundModel: FundWorkflow);
      begin
        fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutdamage', false);
        fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutpopulationperturbation', true);
        fundModel.ModelFields.SetNonDimensionalFieldValue('globalhotellingtaxation','basetax', 0.0);  
      end,
      scenario.eta, scenario.prtp);
      
    Console.WriteLine;
    Console.WriteLine("Without damage");
    Console.WriteLine;
      
    var bgewithoutdamage:=CalculateBGE(
      method(fundModel: FundWorkflow);
      begin
        fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutdamage', true);
        fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutpopulationperturbation', true);
        fundModel.ModelFields.SetNonDimensionalFieldValue('globalhotellingtaxation','basetax', 0.0);  
      end,
      scenario.eta, scenario.prtp);
      
    for each taxresult:KeyValuePair<double,BGE> in OptimumFinder(20,scenario.eta,scenario.prtp) do
    begin
      Console.WriteLine;
      Console.WriteLine("Tax ${0}", taxresult.Key);
      Console.WriteLine;
      
      if not bgebytaxes.ContainsKey(taxresult.Key) then
        bgebytaxes.Add(taxresult.Key,taxresult.Value);
    end;
    
    var calcchange:=method(v1,v2: Double):double;  
        begin
          Result:=v2/v1 - 1.0;
        end;
        
    var changebgeglobal:=calcchange(bgewithdamage.GlobalBGE, bgewithoutdamage.GlobalBGE);
    var changebgeregional:=calcchange(bgewithdamage.RegionalBGE, bgewithoutdamage.RegionalBGE);
    
    BGEResultsCsv.WriteLine('{0};{1};{2};{3};{4};{5};{6};{7};{8};{9};{10};{11};{12};{13};{14};{15};{16};{17};{18};{19};{20};{21}',
      _run.Scenario.Name,
      scenario.prtp,
      scenario.eta,
      'bau',
      bgewithdamage.GlobalBGE,
      bgewithdamage.RegionalBGE,
      0.0,
      0.0,
      bgewithdamage.ExpectedTempIn2050,
      bgewithdamage.ExpectedTempIn2100,
      bgewithdamage.MaxExpectedTemp,
      bgewithdamage.ExpectedRfIn2050,
      bgewithdamage.ExpectedRfIn2100,
      bgewithdamage.MaxExpectedRf,
      bgewithdamage.TrimmedForNaNGlobalWelfare,
      bgewithdamage.TrimmedForNaNRegionalWelfare,
      bgewithdamage.TrimmedForNaNTempIn2050,
      bgewithdamage.TrimmedForNaNTempIn2100,
      bgewithdamage.TrimmedForNaNRfIn2050,
      bgewithdamage.TrimmedForNaNRfIn2100,
      bgewithdamage.TrimmedForNaNTempMax,
      bgewithdamage.TrimmedForNaNRfMax);  
          
      
    BGEResultsCsv.WriteLine('{0};{1};{2};{3};{4};{5};{6};{7};{8};{9};{10};{11};{12};{13};{14};{15};{16};{17};{18};{19};{20};{21}',
      _run.Scenario.Name,
      scenario.prtp,
      scenario.eta,
      'cm',
      bgewithoutdamage.GlobalBGE,
      bgewithoutdamage.RegionalBGE,
      changebgeglobal,
      changebgeregional,
      bgewithoutdamage.ExpectedTempIn2050,
      bgewithoutdamage.ExpectedTempIn2100,
      bgewithoutdamage.MaxExpectedTemp,
      bgewithoutdamage.ExpectedRfIn2050,
      bgewithoutdamage.ExpectedRfIn2100,
      bgewithoutdamage.MaxExpectedRf,
      bgewithoutdamage.TrimmedForNaNGlobalWelfare,
      bgewithoutdamage.TrimmedForNaNRegionalWelfare,
      bgewithoutdamage.TrimmedForNaNTempIn2050,
      bgewithoutdamage.TrimmedForNaNTempIn2100,
      bgewithoutdamage.TrimmedForNaNRfIn2050,
      bgewithoutdamage.TrimmedForNaNRfIn2100,
      bgewithoutdamage.TrimmedForNaNTempMax,
      bgewithoutdamage.TrimmedForNaNRfMax);        
      
    bgeresults.Add(string.Format('prtp: {2:f1}%, Costless improvement: {0:f2}%, regional: {1:f2}%', changebgeglobal*100.0, changebgeregional *100.0, scenario.prtp*100.0));
        
    for each bgebytax in bgebytaxes do
    begin
      changebgeglobal:=calcchange(bgewithdamage.GlobalBGE, bgebytax.Value.GlobalBGE);
      changebgeregional:=calcchange(bgewithdamage.RegionalBGE, bgebytax.Value.RegionalBGE);
      
      BGEResultsCsv.WriteLine('{0};{1};{2};{3};{4};{5};{6};{7};{8};{9};{10};{11};{12};{13};{14};{15};{16};{17};{18};{19};{20};{21}',
        _run.Scenario.Name,
        scenario.prtp,
        scenario.eta,
        bgebytax.Key,
        bgebytax.Value.GlobalBGE,
        bgebytax.Value.RegionalBGE,
        changebgeglobal,
        changebgeregional,
        bgebytax.Value.ExpectedTempIn2050,
        bgebytax.Value.ExpectedTempIn2100,
        bgebytax.Value.MaxExpectedTemp,
        bgebytax.Value.ExpectedRfIn2050,
        bgebytax.Value.ExpectedRfIn2100,
        bgebytax.Value.MaxExpectedRf,
        bgebytax.Value.TrimmedForNaNGlobalWelfare,
        bgebytax.Value.TrimmedForNaNRegionalWelfare,
        bgebytax.Value.TrimmedForNaNTempIn2050,
        bgebytax.Value.TrimmedForNaNTempIn2100,
        bgebytax.Value.TrimmedForNaNRfIn2050,
        bgebytax.Value.TrimmedForNaNRfIn2100,
        bgebytax.Value.TrimmedForNaNTempMax,
        bgebytax.Value.TrimmedForNaNRfMax);  
                                          
      
      bgeresults.Add(string.Format('prtp: {3:f1}%, Tax improvement: {0:f2}, Change in global BGE: {1:f2}%, in regional BGE {2:f2}%', bgebytax.Key, changebgeglobal*100.0, changebgeregional*100.0, scenario.prtp*100.0));    
    end;
  end;

  for each s in bgeresults do
  begin
    Console.WriteLine(s);
  end;  
end;

method TBGERun.CalculateBGE(preRunModifications: ModelModifierDelegate; eta, prtp: double): BGE; 
var
  referenceOutput: ModelOutput;
  bge:=new BGE;  
begin
  _parameters._rand:=new jp.takel.PseudoRandom.MersenneTwister();
  
  _parameters.SetToBestGuessValue;
    
  var randomGlobalWelfareOutcomes := new List<double>;
  var randomRegionalWelfareOutcomes :=new List<double>;
  
  var randomTempIn2050Outcomes := new List<double>;
  var randomTempIn2100Outcomes := new List<double>;
  var randomMaxTempOutcomes := new List<double>;

  var randomRfIn2050Outcomes := new List<double>;
  var randomRfIn2100Outcomes := new List<double>;
  var randomMaxRfOutcomes := new List<double>;    

  for monteCarloRun: Integer := 0 to _run.MonteCarloRuns do
  begin
    if monteCarloRun=0 then
    begin
      Console.WriteLine('Best guess run ');

      _parameters.SetToBestGuessValue;
    end
    else
    begin
      if monteCarloRun mod 20=0 then
      begin
        try
          Console.SetCursorPosition(0, Console.CursorTop - 1);
        except
          on E: IOException do
          begin          
          end;
        end;
        Console.WriteLine('Run {0}                ', monteCarloRun.ToString);
      end;

      _parameters.SetToRandomValue;
    end;
   
    var i_output:=new ModelOutput;
    
    var fundWorkflow:=new FundWorkflow(_parameters); 
    
    // Override parameters
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('globalwelfare','prtp', prtp);
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('globalwelfare','elasticityofmarginalutility', eta);
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','prtp', prtp);
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('regionalwelfare','elasticityofmarginalutility', eta);
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('socioeconomic','prtp', prtp);
    fundWorkflow.ModelFields.SetNonDimensionalFieldValue('socioeconomic','elasticityofmarginalutility', eta);
    
    preRunModifications(fundWorkflow);
    
    fundWorkflow.Run(i_output);  
    
    if monteCarloRun=0 then
      referenceOutput:=i_output;
    
    var globalwelfare:=fundWorkflow.WelfareState.welfare[Timestep.FromSimulationYear(348)];
    var regionalwelfare:=fundWorkflow.RegionalWelfareState.cummulativewelfare[Timestep.FromSimulationYear(348)];
    
    var tempIn2050:=fundWorkflow.ClimateState.temp[Timestep.FromSimulationYear(100)];
    var tempIn2100:=fundWorkflow.ClimateState.temp[Timestep.FromSimulationYear(150)];
    var rfIn2050:=fundWorkflow.ClimateState.radforc[Timestep.FromSimulationYear(100)];
    var rfIn2100:=fundWorkflow.ClimateState.radforc[Timestep.FromSimulationYear(150)];    
    
    var maxTemp:=0.0;
    var maxRf:=0.0;
    var c:=new Esmf.Clock(Timestep.FromSimulationYear(50), Timestep.FromSimulationYear(348));
    
    while not c.IsDone do
    begin
      maxTemp:=Math.Max(maxTemp,fundWorkflow.ClimateState.temp[c.Current]);          
      maxRf:=Math.Max(maxRf,fundWorkflow.ClimateState.radforc[c.Current]); 
      c.Advance;         
    end;
    
    randomGlobalWelfareOutcomes.Add(globalwelfare);
    randomRegionalWelfareOutcomes.Add(regionalwelfare);
    
    randomTempIn2050Outcomes.Add(tempIn2050);
    randomTempIn2100Outcomes.Add(tempIn2100);
    randomRfIn2050Outcomes.Add(rfIn2050);
    randomRfIn2100Outcomes.Add(rfIn2100);
    randomMaxTempOutcomes.Add(maxTemp);
    randomMaxRfOutcomes.Add(maxRf);
  end;
  
  // Remove NaNs
  
  bge.TrimmedForNaNGlobalWelfare  :=randomGlobalWelfareOutcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNRegionalWelfare:=randomRegionalWelfareOutcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNTempIn2050     :=randomTempIn2050Outcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNTempIn2100     :=randomTempIn2100Outcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNRfIn2050       :=randomRfIn2050Outcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNRfIn2100       :=randomRfIn2100Outcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNTempMax        :=randomMaxTempOutcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));
  bge.TrimmedForNaNRfMax          :=randomMaxRfOutcomes.RemoveAll(d->Double.IsNaN(d) or Double.IsInfinity(d));  
  
  // Sort
  
  randomGlobalWelfareOutcomes.Sort();
  randomRegionalWelfareOutcomes.Sort();
  randomTempIn2050Outcomes.Sort();
  randomTempIn2100Outcomes.Sort();
  randomRfIn2050Outcomes.Sort();
  randomRfIn2100Outcomes.Sort();
  randomMaxTempOutcomes.Sort();
  randomMaxRfOutcomes.Sort();
   
  // Take out top and bottom 10
  var percenttrim:=0.01;
  
  var numtrim:=integer(Math.Floor(randomGlobalWelfareOutcomes.Count*percenttrim));    
  
  randomGlobalWelfareOutcomes.RemoveRange(0,numtrim);
  randomGlobalWelfareOutcomes.RemoveRange(randomGlobalWelfareOutcomes.Count-numtrim, numtrim);
  
  randomRegionalWelfareOutcomes.RemoveRange(0,numtrim);
  randomRegionalWelfareOutcomes.RemoveRange(randomRegionalWelfareOutcomes.Count-numtrim, numtrim); 
  
  randomTempIn2050Outcomes.RemoveRange(0,numtrim);
  randomTempIn2050Outcomes.RemoveRange(randomTempIn2050Outcomes.Count-numtrim, numtrim); 
  
  randomTempIn2100Outcomes.RemoveRange(0,numtrim);
  randomTempIn2100Outcomes.RemoveRange(randomTempIn2100Outcomes.Count-numtrim, numtrim); 
  
  randomRfIn2050Outcomes.RemoveRange(0,numtrim);
  randomRfIn2050Outcomes.RemoveRange(randomRfIn2050Outcomes.Count-numtrim, numtrim); 
  
  randomRfIn2100Outcomes.RemoveRange(0,numtrim);
  randomRfIn2100Outcomes.RemoveRange(randomRfIn2100Outcomes.Count-numtrim, numtrim); 
  
  randomMaxTempOutcomes.RemoveRange(0,numtrim);
  randomMaxTempOutcomes.RemoveRange(randomMaxTempOutcomes.Count-numtrim, numtrim); 
  
  randomMaxRfOutcomes.RemoveRange(0,numtrim);
  randomMaxRfOutcomes.RemoveRange(randomMaxRfOutcomes.Count-numtrim, numtrim);   
  
  
//  Console.WriteLine('Left over for global: {0} and regional: {1}', randomGlobalWelfareOutcomes.Count,
//    randomRegionalWelfareOutcomes.Count);

  bge.ExpectedRfIn2050:=0.0;
  bge.ExpectedRfIn2100:=0.0;
  bge.ExpectedTempIn2050:=0.0;
  bge.ExpectedTempIn2100:=0.0;
  bge.MaxExpectedRf:=0.0;
  bge.MaxExpectedTemp:=0.0;
  
  for each d in randomTempIn2050Outcomes do
  begin
    bge.ExpectedTempIn2050:=bge.ExpectedTempIn2050 + d;
  end;
  bge.ExpectedTempIn2050:=bge.ExpectedTempIn2050 / double(randomTempIn2050Outcomes.Count);
  
  for each d in randomTempIn2100Outcomes do
  begin
    bge.ExpectedTempIn2100:=bge.ExpectedTempIn2100 + d;
  end;
  bge.ExpectedTempIn2100:=bge.ExpectedTempIn2100 / double(randomTempIn2100Outcomes.Count);

  for each d in randomMaxTempOutcomes do
  begin
    bge.MaxExpectedTemp:=bge.MaxExpectedTemp + d;
  end;
  bge.MaxExpectedTemp:=bge.MaxExpectedTemp / double(randomMaxTempOutcomes.Count);

  for each d in randomRfIn2050Outcomes do
  begin
    bge.ExpectedRfIn2050:=bge.ExpectedRfIn2050 + d;
  end;
  bge.ExpectedRfIn2050:=bge.ExpectedRfIn2050 / double(randomRfIn2050Outcomes.Count);
  
  for each d in randomRfIn2100Outcomes do
  begin
    bge.ExpectedRfIn2100:=bge.ExpectedRfIn2100 + d;
  end;
  bge.ExpectedRfIn2100:=bge.ExpectedRfIn2100 / double(randomRfIn2100Outcomes.Count);

  for each d in randomMaxRfOutcomes do
  begin
    bge.MaxExpectedRf:=bge.MaxExpectedRf + d;
  end;
  bge.MaxExpectedRf:=bge.MaxExpectedRf / double(randomMaxRfOutcomes.Count);  


  
  var expectedGlobalWelfare:=0.0;
  var expectedRegionalWelfare:=0.0;
  
  for each w in randomGlobalWelfareOutcomes do
  begin
    expectedGlobalWelfare:=expectedGlobalWelfare + w;
  end;
  expectedGlobalWelfare:=expectedGlobalWelfare / double(randomGlobalWelfareOutcomes.Count);
  
  for each w in randomRegionalWelfareOutcomes do
  begin
    expectedRegionalWelfare:=expectedRegionalWelfare + w;
  end;
  expectedRegionalWelfare:=expectedRegionalWelfare / double(randomRegionalWelfareOutcomes.Count); 
       
  //Console.WriteLine('Expected welfare global: {0}, regional: {1}', expectedGlobalWelfare, expectedRegionalWelfare);
  var gamma:=0.02;
  
  if eta=1.0 then
  begin    
    var alpha:=0.0;  
    for i:Integer := 50 to 348 do
    begin
      var t:double := double(i)-50.0;
      
      var population: Double :=0;
      for l: Integer :=1 to LegacyConstants.NoReg do
        population:=population + referenceOutput.Populations.Item[i,l];
        
      alpha:=alpha + Math.Log(Math.Pow(1.0 + gamma,t))*population*Math.Pow(1.0+prtp, -t);
    end;
    
    var beta:=0.0;  
    for i:Integer := 50 to 348 do
    begin
      var t:double := double(i)-50.0;
      
      var population: Double :=0;
      for l: Integer :=1 to LegacyConstants.NoReg do
        population:=population + referenceOutput.Populations.Item[i,l];
        
      beta:=beta + population*Math.Pow(1.0+prtp, -t);
    end;
    
    bge.GlobalBGE:= Math.Exp((expectedGlobalWelfare - alpha)/beta);
    bge.RegionalBGE:= Math.Exp((expectedRegionalWelfare - alpha)/beta);
  end
  else
  begin
    var alpha:=0.0;  
    for i:Integer := 50 to 348 do
    begin
      var t:double := double(i)-50.0;
      
      var population: Double :=0;
      for l: Integer :=1 to LegacyConstants.NoReg do
        population:=population + referenceOutput.Populations.Item[i,l];
        
      alpha:=alpha + Math.Pow(1.0 + gamma, t*(1.0-eta))/(1.0-eta)*population*Math.Pow(1.0+prtp, -t);
    end;
    
    bge.GlobalBGE:=Math.Pow(expectedGlobalWelfare/alpha, 1.0/(1.0-eta));
    bge.RegionalBGE:=Math.Pow(expectedRegionalWelfare/alpha, 1.0/(1.0-eta));
  end;

//  Console.WriteLine('BGE global: {0}, regional: {1}', bge.GlobalBGE, bge.RegionalBGE);

  Result:=bge;
end;

method TBGERun.OptimumFinder(maxNumberOfSteps: Integer; eta, prtp: double): IEnumerable<KeyValuePair<double,BGE>>; 
begin
  var stepcounter:=0;
  var stepsize:=20.0;
  var currenttax:=0.0;
  var phaseone:=true;
  
  var swf:=SWFtoOptimize(_run.SWFToOptimise);
  
  var swfswitcher:=method(v: BGE): Double;
    begin      
      case swf of
        SWFtoOptimize.GlobalSwf: result:=v.GlobalBGE;
        SWFtoOptimize.RegionalSwf: result:=v.RegionalBGE;
        else raise new Exception('Unknown SWFToOptimise value');
      end;
    end;

  var lastbge:=CalculateBGE(
    method(fundModel: FundWorkflow);
    begin
      fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutdamage', false);
      fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutpopulationperturbation', true);
      fundModel.ModelFields.SetNonDimensionalFieldValue('globalhotellingtaxation','basetax', currenttax);
    end,
    eta, prtp);
    
  var notaxationresult:=swfswitcher(lastbge);
    
  yield(new KeyValuePair<double,BGE>(currenttax,lastbge));    
  
  currenttax:=+stepsize;
    
  while (stepcounter<maxNumberOfSteps) and (Math.Abs(stepsize)>0.5) do
  begin
    var bge:=CalculateBGE(
        method(fundModel: FundWorkflow);
        begin
          fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutdamage', false);
          fundModel.ModelFields.SetNonDimensionalFieldValue('socioeconomic','runwithoutpopulationperturbation', true);
          fundModel.ModelFields.SetNonDimensionalFieldValue('globalhotellingtaxation','basetax', currenttax);
        end,
        eta, prtp);
    
    yield(new KeyValuePair<double,BGE>(currenttax,bge));
    
    if phaseone and (swfswitcher(bge)<notaxationresult) then
    begin
      currenttax:=currenttax-stepsize;
      stepsize:=stepsize/2.0;
      bge:=lastbge;
    end        
    else if swfswitcher(bge)<swfswitcher(lastbge) then
    begin
      phaseone:=false;
      
      stepsize:=stepsize/(-2.0);
    end
    else
      phaseone:=false;
    
    currenttax:=currenttax + stepsize;
    
    if currenttax<=0.0 then
    begin
      stepsize:=stepsize/(-2.0);
      currenttax:=stepsize;
    end;
    
    lastbge:=bge;
    inc(stepcounter);
  end;
  
  Console.WriteLine('{0} steps used', stepcounter);
  
end;

end.
