// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.3
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
namespace Fund;

interface

type

  // Discounting with a constant pure rate of time preference
  ConstantDiscounting = public class (WeightingScheme)
  private
    m_pureRateOfTimePreference: double;
    m_riskAversion: Double;
    m_baseYear: integer;
    m_name: String;
  protected
    function GetName: string; override;
  public
    constructor(PureRateOfTimePreference: double);
    constructor(PureRateOfTimePreference: double; baseYear: integer);
    constructor(PureRateOfTimePreference: Double; RiskAversion: double; baseYear: Integer);
    constructor(PureRateOfTimePreference: double; baseYear: integer; name: String);

    procedure CalculateWeights(ModelOutput: ModelOutput); override;
  end; 
  
implementation

procedure ConstantDiscounting.CalculateWeights(ModelOutput: ModelOutput);
var
  i_newPerCapitaIncome, i_oldPerCapitaIncome: double;
  i_perCapitaGrowthRate: double;
  i_discountFactor: double;
  i, l: Integer;
begin
  inherited;
  {
  i_discountFactor:=1;
  for i:=50 to LegacyConstants.NYear do
  begin
    for l:=1 to LegacyConstants.NoReg do
      AddWeight(i, l, i_discountFactor);

    i_totalNewIncome:=0;
    i_totalNewPopulation:=0;
    i_totalOldIncome:=0;
    i_totalOldPopulation:=0;
    for l:=1 to LegacyConstants.NoReg do
    begin
      i_totalNewIncome:=i_totalNewIncome + ModelOutput.Incomes.Item[i, l];
      i_totalNewPopulation:=i_totalNewPopulation + ModelOutput.Populations.Item[i, l];

      i_totalOldIncome:=i_totalOldIncome + ModelOutput.Incomes.Item[i-1, l];
      i_totalOldPopulation:=i_totalOldPopulation + ModelOutput.Populations.Item[i-1, l];
    end;
    i_newPerCapitaIncome:=i_totalNewIncome/i_totalNewPopulation;
    i_oldPerCapitaIncome:=i_totalOldIncome/i_totalOldPopulation;
    i_perCapitaGrowthRate:=(i_newPerCapitaIncome-i_oldPerCapitaIncome)/i_oldPerCapitaIncome;
    i_discountFactor:=i_discountFactor/(1.00 + m_PureRateOfTimePreference + i_perCapitaGrowthRate);
  end;  }

  for i:=1 to LegacyConstants.NoReg do
  begin
    i_discountFactor:=1;
    // TODO -oDavid Anthoff: Change start of "real" period to constant
    for l:=m_baseYear to LegacyConstants.NYear do
    begin
      AddWeight(l, i, i_discountFactor);
      i_newPerCapitaIncome:=ModelOutput.Incomes.Item[l, i]/ModelOutput.Populations.Item[l, i];
      i_oldPerCapitaIncome:=ModelOutput.Incomes.Item[l-1, i]/ModelOutput.Populations.Item[l-1, i];
      i_perCapitaGrowthRate:=(i_newPerCapitaIncome-i_oldPerCapitaIncome)/i_oldPerCapitaIncome;
      i_discountFactor:=i_discountFactor/(1.00 + m_PureRateOfTimePreference + m_riskAversion * i_perCapitaGrowthRate);
    end;
  end;

end;

constructor ConstantDiscounting(PureRateOfTimePreference: double);
begin
  inherited constructor;

  m_pureRateOfTimePreference:=PureRateOfTimePreference;
  m_baseYear:=50;
  m_riskAversion:=1.0;
  m_name:=String.Format('constant discount rate [prtp={0}%, risk aversion={1}, base year={2}]', m_pureRateOfTimePreference*100, m_riskAversion, m_baseYear+1950);
end;

constructor ConstantDiscounting(PureRateOfTimePreference: double; baseYear: integer);
begin
  inherited constructor;
  
  m_pureRateOfTimePreference:=PureRateOfTimePreference;
  m_baseYear:=baseYear;
  m_riskAversion:=1.0;
  m_name:=String.Format('constant discount rate [prtp={0}%, risk aversion={1}, base year={2}]', m_pureRateOfTimePreference*100, m_riskAversion, m_baseYear+1950);
end;

function ConstantDiscounting.GetName: string;
begin
  Result:=m_name;
end;

constructor ConstantDiscounting(PureRateOfTimePreference: double; baseYear: integer; name: String); 
begin
  inherited constructor;
  
  m_pureRateOfTimePreference:=PureRateOfTimePreference;
  m_baseYear:=baseYear;
  m_name:=name;
  m_riskAversion:=1.0;
end;

constructor ConstantDiscounting(PureRateOfTimePreference: Double; RiskAversion: double; baseYear: Integer); 
begin
  m_pureRateOfTimePreference:=PureRateOfTimePreference;
  m_riskAversion:=RiskAversion;
  m_baseYear:=baseYear;
  m_name:=String.Format('constant discount rate [prtp={0}%, risk aversion={1}, base year={2}]', m_pureRateOfTimePreference*100, m_riskAversion, m_baseYear+1950);
end;

end.